import os
from warnings import warn
from mewpy.regulation import RegulatoryModel, RegulatoryVariable, RegulatoryInteraction
from mewpy.io import read_tabular_aliases, read_tabular_regulatory_model
from mewpy.regulation.regulatory_model import solution_decode
from mewpy.simulation import get_simulator, SStatus
from mewpy.simulation import SimulationMethod
from mewpy.simulation.cobra import Simulation as Cobra_model
from mewpy.simulation.reframed import Simulation as Reframed_model
from mewpy.simulation.simulation import SimulationResult
from mewpy.utils.parsing import boolean_rule_from_str, BooleanEvaluator


class IntegratedRegulatoryVariable(RegulatoryVariable):
    """
    Integrated Regulatory Variable

    """

    def __init__(self, identifier,
                 name=None,
                 expression_coef=None,
                 aliases=None,
                 regulatory_interactions=None,
                 targets=None,
                 regulators=None,
                 model=None,
                 cbm_model=None,
                 symbol=None,
                 is_regulator=False,
                 is_target=False):

        self._symbol = None
        self._cbm_model = None
        self._cbm_model_id = None

        super().__init__(identifier,
                         name=name,
                         expression_coef=expression_coef,
                         aliases=aliases,
                         regulatory_interactions=regulatory_interactions,
                         targets=targets,
                         regulators=regulators,
                         model=model,
                         is_regulator=is_regulator,
                         is_target=is_target)

        self._build_integrated_regulatory_variable(cbm_model, symbol)

    def _build_integrated_regulatory_variable(self, cbm_model, symbol):

        if symbol is not None:
            self.symbol = symbol
        self.cbm_model = cbm_model

    @property
    def id(self):
        return getattr(self, '_id', None)

    @property
    def aliases(self):
        return getattr(self, '_aliases', None)

    @property
    def symbol(self):
        return getattr(self, '_symbol', None)

    @property
    def cbm_model(self):
        return getattr(self, '_cbm_model', None)

    @property
    def cbm_model_id(self):
        return getattr(self, '_cbm_model_id', None)

    @id.setter
    def id(self, value):

        if value == self._id:
            pass

        elif not isinstance(value, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        else:

            # symbol, alias = variable_from_str(value)
            #
            # self.symbol = symbol
            # self.aliases = list(alias.values())

            # Uncomment this if we want to keep available the option for altering an id of a given variable

            # if self.model:
            #
            #     if value in self.model.regulators and value in self.model.targets:
            #         def reg_id(message):
            #             warn(message, UserWarning, stacklevel=2)
            #
            #         reg_id("The ID already exists in the dictionary of regulators and targets of the regulatory "
            #                "model. Ignoring the alteration")
            #
            #     else:
            #
            #         if self._id in self.model.regulators:
            #             del self.model.regulators[self._id]
            #             self.model.regulators[value] = self
            #
            #         if self._id in self.model.targets:
            #             del self.model.targets[self._id]
            #             self.model.targets[value] = self
            #
            #         if self.regulatory_interactions:
            #             for reg_inter in self.regulatory_interactions_gen():
            #
            #                 if self._id in reg_inter.regulators:
            #                     del reg_inter.regulators[self._id]
            #                     reg_inter.regulators[value] = self
            #
            #         if self.regulators:
            #             for reg in self.regulators.values():
            #                 del reg.targets[self._id]
            #                 reg.targets[value] = self
            #
            #         if self.targets:
            #             for tar in self.targets.values():
            #                 del tar.regulators[self._id]
            #                 tar.regulators[value] = self
            #
            #         self._id = value
            #
            # else:
            #
            #     if self.regulatory_interactions:
            #
            #         for reg_inter in self.regulatory_interactions_gen():
            #
            #             if self._id in reg_inter.regulators:
            #                 del reg_inter.regulators[self._id]
            #                 reg_inter.regulators[value] = self

            self._id = value

    @aliases.setter
    def aliases(self, value):

        if value == self._aliases:
            pass

        elif not value:
            self._aliases = []

        else:

            if not isinstance(value, list):
                raise TypeError("The aliases must be a list. Use list() to convert")

            if value and not isinstance(value[0], str):
                raise TypeError("The aliases must be a list. Use list() to convert")

            if not self._aliases:
                self._aliases = []

            for val in value:
                if val not in self._aliases:
                    # Reframed keeps the G_, M_ and R_ in sbml files
                    self._aliases = self._aliases + [val, 'G_' + val, 'M_' + val, 'R_' + val]

    @symbol.setter
    def symbol(self, value):

        from sympy.core.symbol import Symbol

        if value == self._symbol:

            pass

        elif not value:

            self._symbol = None

        elif not isinstance(value, Symbol):

            raise TypeError("The symbol must be a sympy.core.symbol.Symbol")

        else:

            self._symbol = value

    @cbm_model.setter
    def cbm_model(self, value):

        if value == self._cbm_model:
            pass

        elif not value:
            self._cbm_model = None

        else:

            if not isinstance(value, Reframed_model) and not isinstance(value, Cobra_model):

                self._cbm_model = get_simulator(value)

            else:

                self._cbm_model = value

            if self.id in self._cbm_model.genes:
                self._cbm_model_id = self.id

            elif self.id in self._cbm_model.reactions:
                self._cbm_model_id = self.id

            elif self.id in self._cbm_model.metabolites:
                self._cbm_model_id = self.id

            else:

                for alias in self.aliases:
                    if alias in self._cbm_model.genes:
                        self._cbm_model_id = alias
                        break
                    if alias in self._cbm_model.reactions:
                        self._cbm_model_id = alias
                        break
                    if alias in self._cbm_model.metabolites:
                        self._cbm_model_id = alias
                        break

    def ko(self):

        """
        Knock-out a given regulatory variable means setting the expression coefficient to zero
        :return:
        """

        self.expression_coef = 0


class IntegratedRegulatoryInteraction(RegulatoryInteraction):
    """
    Integrated Regulatory Interaction

    """

    def __init__(self, identifier,
                 rule,
                 target,
                 regulators=None,
                 model=None,
                 aliases=None):

        self._conditionals = []
        self._variables = None
        self._tree = None

        super().__init__(identifier,
                         rule,
                         target,
                         regulators=regulators,
                         model=model,
                         aliases=aliases)

    @classmethod
    def from_str(cls, identifier,
                 rule,
                 target,
                 target_aliases=None,
                 regulators=None,
                 model=None,
                 cbm_model=None):

        if not isinstance(identifier, str):
            raise TypeError("IDs must be a str object")

        if not isinstance(rule, str):
            raise TypeError("rule must be a string object")

        if isinstance(target, str):

            if model is not None:

                tar_var = model.add_target(identifier=target,
                                           name=target,
                                           aliases=target_aliases,
                                           regulators=regulators)

            else:

                tar_var = IntegratedRegulatoryVariable(identifier=target,
                                                       name=target,
                                                       aliases=target_aliases,
                                                       cbm_model=cbm_model)

            return cls(identifier=identifier,
                       rule=rule,
                       target=tar_var,
                       model=model)

        elif isinstance(target, IntegratedRegulatoryVariable):

            return cls(identifier=identifier,
                       rule=rule,
                       target=target,
                       model=model)

        else:
            raise TypeError("Target must be a string or a RegulatoryVariable object")

    @property
    def id(self):
        return getattr(self, '_id', None)

    @property
    def rule(self):
        return getattr(self, '_rule', None)

    # @property
    # def sympify(self):
    #     return getattr(self, '_sympify', None)
    #
    # @property
    # def lambdify(self):
    #     return getattr(self, '_lambdify', None)

    @property
    def variables(self):
        return getattr(self, '_variables', None)

    @property
    def tree(self):
        return getattr(self, '_tree', None)

    @id.setter
    def id(self, value):

        if value == self._id:
            pass

        elif not isinstance(value, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        else:

            # Uncomment this if we want to keep available the option for altering an id of a given variable
            # change for the regulatory variables property

            # if self.model:
            #
            #     if value in self.model.regulators and value in self.model.targets:
            #         def reg_id(message):
            #             warn(message, UserWarning, stacklevel=2)
            #
            #         reg_id("The ID already exists in the dictionary of regulators and targets of the regulatory "
            #                "model. Ignoring the alteration")
            #
            #     else:
            #
            #         if self._id in self.model.regulatory_interactions:
            #             del self.model.regulatory_interactions[self._id]
            #             self.model.regulatory_interactions[value] = self
            #
            #         if self.regulators:
            #             for reg in self.regulators_gen():
            #                 if self._id in reg.regulatory_interactions:
            #                     del reg.regulatory_interactions[self._id]
            #                     reg.regulatory_interactions[value] = self
            #
            #         if self.target:
            #             if self._id in self.target.regulatory_interactions:
            #                 del self.target.regulatory_interactions[self._id]
            #                 self.target.regulatory_interactions.regulatory_interactions[value] = self
            #
            #         self._id = value
            #
            # else:
            #
            #     if self.regulators:
            #         for reg in self.regulators_gen():
            #             if self._id in reg.regulatory_interactions:
            #                 del reg.regulatory_interactions[self._id]
            #                 reg.regulatory_interactions[value] = self
            #
            #     if self.target:
            #         if self._id in self.target.regulatory_interactions:
            #             del self.target.regulatory_interactions[self._id]
            #             self.target.regulatory_interactions.regulatory_interactions[value] = self

            self._id = value

    @rule.setter
    def rule(self, value):

        if value == self._rule:
            pass

        elif not isinstance(value, str):
            raise TypeError("Regulatory interaction rule must be a string. Use str() to convert")

        else:

            self.old_rule = value

            self._rule, self._elements, self._tree, self._variables, aliases, self._conditionals = \
                boolean_rule_from_str(value)

            if self.regulators:

                for key in self.regulators:
                    if key not in self.variables:
                        raise KeyError("The regulators property doesn't match all regulators in the regulatory "
                                       "interaction rule")

                    self.regulators[key].regulatory_interactions[self.id] = self
                    self.regulators[key].is_regulator = True
                    self.regulators[key].model = self.model
                    self.regulators[key].targets[self.target.id] = self.target
                    self.regulators[key].aliases = aliases[key]
                    # self.regulators[key].symbol = symbols[key]

                    self.target.regulators[key] = self.regulators[key]

            else:

                for reg in self.variables:

                    if self.model is not None:

                        reg_var = self.model.add_regulator(identifier=reg,
                                                           name=reg,
                                                           aliases=aliases[reg],
                                                           regulatory_interactions={self.id: self},
                                                           targets={self.target.id: self.target})

                    else:

                        reg_var = IntegratedRegulatoryVariable(identifier=reg,
                                                               name=reg,
                                                               aliases=aliases[reg],
                                                               regulatory_interactions={self.id: self},
                                                               targets={self.target.id: self.target})

                    self.regulators[reg] = reg_var
                    self.target.regulators[reg] = reg_var

    # @sympify.setter
    # def sympify(self, value):
    #
    #     raise BaseException("lambdify property cannot be changed. "
    #                         "please create/set a new regulatory interaction rule.")
    #
    # @lambdify.setter
    # def lambdify(self, value):
    #
    #     raise BaseException("lambdify property cannot be changed. "
    #                         "please create/set a new regulatory interaction rule.")

    @variables.setter
    def variables(self, value):

        raise BaseException("variables property cannot be changed. "
                            "please create/set a new regulatory interaction rule.")

    # @tree.setter
    # def tree(self, value):
    #
    #     from mewpy.utils.parsing import Node
    #
    #     if value == self._tree:
    #         pass
    #
    #     if not isinstance(value, Node):
    #         raise TypeError("The tree must be a Node object. Use Node() to convert")
    #
    #     else:
    #         self._tree = value

    @rule.deleter
    def rule(self):
        del self._rule
        del self._variables
        del self._tree

    def evaluate(self, state_map=None):

        if self.tree.value is None:
            return

        if not state_map:
            state_map = {}

        true_list = []
        _state_map = {reg.id: reg.expression_coef for reg in self.regulators.values()}

        new_state_map = {}
        for key, value in _state_map.items():

            if key in state_map:
                value = state_map[key]

            if key not in self._conditionals:

                if value:
                    true_list.append(key)

            else:
                new_state_map[key] = value

        evaluator = BooleanEvaluator(true_list, new_state_map)
        res = self.tree.evaluate(evaluator.f_operand, evaluator.f_operator)
        res = solution_decode(res)

        # print("{} evaluation result: {}".format(self.id, str(res)))

        return res

    def print_tree(self):

        self.tree.print_node()


class IntegratedModel(RegulatoryModel):
    """
    Integrated Model base class.
    Do not instantiate this class. This works as bridge/interface for linking metabolic and regulatory entities found
    in regulatory and cbm models. This is crucial for the integrated methods analysis available at mewpy (e.g. RFBA and
    SRFBA) to work properly.

    Notes:
        -Only supports boolean regulatory models (by now)

    """

    def __init__(self,
                 identifier,
                 name=None,
                 cbm_model=None,
                 cbm_simulation_interface=None,
                 targets=None,
                 regulators=None,
                 regulatory_interactions=None,
                 initial_state=None):

        # CBM model
        self._cbm_model = None
        self._cbm_simulation_interface = None

        # New from CBM model
        self._cbm_simulation_method = SimulationMethod.pFBA
        self._maximize = True
        self._gprs = {}
        self._gprs_evaluator = {}
        self._metabolic_regulatory_genes = {}
        self._metabolic_regulatory_metabolites = {}
        self._metabolic_regulatory_reactions = {}
        self._regulatory_conditions = {}
        self._aliases_map = {}
        # A self._metabolic_regulatory_gprs can be implemented by replacing a given gene into its gpr rules by the
        # regulatory rule associated

        # Regulatory model
        super().__init__(identifier,
                         name,
                         targets,
                         regulators,
                         regulatory_interactions)

        self._initial_state = {}
        self._solution = []

        # Build
        self.__build_integrated_model__(cbm_model, cbm_simulation_interface, initial_state)

    def __build_integrated_model__(self, cbm_model, cbm_simulation_interface, initial_state):

        if cbm_simulation_interface is not None:

            if cbm_model is None:
                raise BaseException("A REFRAMED or COBRApy Constraint-based model must be provided")

            else:
                self._cbm_model = cbm_model

            self.cbm_simulation_interface = cbm_simulation_interface

        else:

            self.cbm_model = cbm_model

        self.initial_state = initial_state

        self.__infer_metabolic_regulatory_variables__()

    @classmethod
    def from_tabular_format(cls, regulatory_file,
                            cbm_model,
                            cbm_simulation_interface=None,
                            identifier=None,
                            initial_state=None,
                            filter_nan=False,
                            sep=None,
                            id_col=None,
                            rule_col=None,
                            aliases_cols=None,
                            header=None,
                            special_chars=None,
                            compartments_ids=None):

        """

        The standard class method for loading an Integrated Model of any type from a tabular format file (txt, csv, etc)
        This method will build an integrated model from scratch using a cobra CBM model, reframed CBM model,
        or mewpy simulation object, and the regulatory rules written in tabular-like format file.

        NOTE THAT:
            -Only one rule is allowed per target.
            -So for only boolean rules are supported
            -Special chars present in the regulatory variables and regulatory rules will be filter and replaced by char
            description plus double underscore (e.g. / -> _slash_).
            -Regulatory variables that start with digits will also be filter and replaced

        Rules can be set as such:

            G1, FruE, G2 and (G3 or M_acetate>0)
            G2, FruR, ON
            G3, AcR, M_02>5

        :param regulatory_file: str, path to the regulatory model
        :param cbm_model: cobra or reframed model objects
        :param cbm_simulation_interface: mewpy simulation object
        :param identifier: str, integrated model identifier
        :param initial_state: dict, initial state of the regulatory variables. If none, the state of all regulatory variables is set to off (zero)
        :param filter_nan: bool, filter targets with empty rules
        :param sep: str, separator
        :param id_col: int, index of the column of the regulatory variables identifiers (those that appear in the regulatory rules)
        :param rule_col: int, index of the column of the regulatory rules
        :param aliases_cols: int, index of the column of the regulatory variables aliases
        :param header: int or None, index of the header row or None if there is none
        :param special_chars: dict, key is the special char, value is its replacement
        :param compartments_ids: dict, key is the compartment id, value is its replacement
        :return:
        """

        if not isinstance(regulatory_file, str):
            raise TypeError("File or file path must be a string")

        if not identifier:
            _, identifier = os.path.split(regulatory_file)
            identifier = os.path.splitext(identifier)[0]

        model = cls(identifier=identifier,
                    name=identifier,
                    cbm_model=cbm_model,
                    cbm_simulation_interface=cbm_simulation_interface,
                    initial_state=None)

        df = read_tabular_regulatory_model(regulatory_file, sep=sep, id_col=id_col, rule_col=rule_col,
                                           aliases_cols=aliases_cols, header=header, special_chars=special_chars,
                                           compartments_ids=compartments_ids)

        aliases_cols = [col for col in df.columns if col != 'ids' and col != 'rules']

        for gene in df.index:

            reg_inter_id = gene + '_regulation'
            reg_inter_rule = df.loc[gene, 'rules']
            target_aliases = list(df.loc[gene, aliases_cols])

            if filter_nan and str(reg_inter_rule) == '':
                continue

            model.add_regulatory_interaction(reg_inter_id, reg_inter_rule, gene, target_aliases=target_aliases)

        # last part of build integrated model, as when the initializer is ran, the regulatory variables are not yet
        # there
        model.initial_state = initial_state

        model.__infer_metabolic_regulatory_variables__()

        return model

    @property
    def cbm_model(self):
        return getattr(self, '_cbm_model', None)

    @property
    def cbm_simulation_interface(self):
        return getattr(self, '_cbm_simulation_interface', None)

    @property
    def reactions(self):
        return self.cbm_simulation_interface.reactions

    @property
    def genes(self):
        return self.cbm_simulation_interface.genes

    @property
    def metabolites(self):
        return self.cbm_simulation_interface.metabolites

    @property
    def medium(self):
        return self.cbm_simulation_interface.medium

    @property
    def objective(self):
        return self.cbm_simulation_interface.objective

    @property
    def environmental_conditions(self):
        return self.cbm_simulation_interface.environmental_conditions

    @property
    def constraints(self):
        return self.cbm_simulation_interface.constraints

    @property
    def solver(self):
        return self.cbm_simulation_interface.solver

    @property
    def reference(self):
        return self.cbm_simulation_interface.reference

    @property
    def cbm_simulation_method(self):
        return getattr(self, '_cbm_simulation_method', None)

    @property
    def maximize(self):
        return getattr(self, '_maximize', None)

    @property
    def gprs(self):
        return getattr(self, '_gprs', None)

    @property
    def metabolic_regulatory_genes(self):
        return getattr(self, '_metabolic_regulatory_genes', None)

    @property
    def metabolic_regulatory_metabolites(self):
        return getattr(self, '_metabolic_regulatory_metabolites', None)

    @property
    def metabolic_regulatory_reactions(self):
        return getattr(self, '_metabolic_regulatory_reactions', None)

    @property
    def regulatory_conditions(self):
        return getattr(self, '_regulatory_conditions', None)

    @property
    def initial_state(self):
        return getattr(self, '_initial_state', None)

    @property
    def solution(self):
        return getattr(self, '_solution', None)

    # @property
    # def metabolic_regulatory_gprs(self):
    #     return getattr(self, '_metabolic_regulatory_gprs', None)

    @cbm_model.setter
    def cbm_model(self, value):

        if value == self._cbm_model:
            pass

        elif not value:
            raise BaseException("A REFRAMED or COBRApy Constraint-based model must be provided")

        else:

            self.cbm_simulation_interface = get_simulator(value)

            self._cbm_model = value

    @cbm_simulation_interface.setter
    def cbm_simulation_interface(self, value):

        if value == self._cbm_simulation_interface:
            pass

        elif not value:
            raise BaseException("A simulation object must be provided")

        elif not isinstance(value, Reframed_model) and not isinstance(value, Cobra_model):
            raise TypeError("A simulation object must be provided")

        else:

            self._cbm_simulation_interface = value

            # Building utils
            self.__build_gprs__()
            self.__infer_metabolic_regulatory_variables__()

    @reactions.setter
    def reactions(self, value):

        raise BaseException("reactions property cannot be changed")

    @genes.setter
    def genes(self, value):

        raise BaseException("genes property cannot be changed")

    @metabolites.setter
    def metabolites(self, value):

        raise BaseException("metabolites property cannot be changed")

    @medium.setter
    def medium(self, value):

        raise BaseException("medium property cannot be changed")

    @objective.setter
    def objective(self, value):

        if value == self._cbm_simulation_interface.objective:
            pass

        elif not value:
            pass

        elif not isinstance(value, dict):
            raise TypeError("The objective must be a dict object")

        else:

            self._cbm_simulation_interface.objective = value

    @environmental_conditions.setter
    def environmental_conditions(self, value):

        if value == self._cbm_simulation_interface.environmental_conditions:
            pass

        elif not value:
            self._cbm_simulation_interface.environmental_conditions = {}

        elif not isinstance(value, dict):
            raise TypeError("The environmental conditions must be a dict object")

        else:

            self._cbm_simulation_interface.environmental_conditions = value

    @constraints.setter
    def constraints(self, value):

        if value == self._cbm_simulation_interface.constraints:
            pass

        elif not value:
            self._cbm_simulation_interface.constraints = {}

        elif not isinstance(value, dict):
            raise TypeError("The constraints must be a dict object")

        else:

            self._cbm_simulation_interface.constraints = value

    @solver.setter
    def solver(self, value):

        if value == self._cbm_simulation_interface.solver:
            pass

        elif not value:
            pass

        elif not isinstance(value, str):
            raise TypeError("The solver must be a str object")

        else:

            self._cbm_simulation_interface.solver = value

    @reference.setter
    def reference(self, value):

        raise BaseException("Reference property cannot be changed")

    @cbm_simulation_method.setter
    def cbm_simulation_method(self, value):

        if value == self._cbm_simulation_method:
            pass

        elif not value:
            pass

        else:
            self._cbm_simulation_method = value

    @maximize.setter
    def maximize(self, value):

        if value == self._maximize:
            pass

        elif value is None:
            pass

        elif isinstance(value, bool):
            raise TypeError("Maximize property must be a boolean object")

        else:
            self._maximize = value

    @gprs.setter
    def gprs(self, value):

        if value == self._gprs:
            pass

        else:
            raise BaseException("gprs property cannot be changed")

    @metabolic_regulatory_genes.setter
    def metabolic_regulatory_genes(self, value):

        if value == self._metabolic_regulatory_genes:
            pass

        else:
            raise BaseException("metabolic_regulatory_genes property cannot be changed")

    @metabolic_regulatory_metabolites.setter
    def metabolic_regulatory_metabolites(self, value):

        if value == self._metabolic_regulatory_metabolites:
            pass

        else:
            raise BaseException("metabolic_regulatory_metabolites property cannot be changed")

    @metabolic_regulatory_reactions.setter
    def metabolic_regulatory_reactions(self, value):

        if value == self._metabolic_regulatory_reactions:
            pass

        else:
            raise BaseException("metabolic_regulatory_reactions property cannot be changed")

    @regulatory_conditions.setter
    def regulatory_conditions(self, value):

        if value == self._regulatory_conditions:
            pass

        else:
            raise BaseException("regulators_environmental_conditions property cannot be changed")

    @initial_state.setter
    def initial_state(self, value):

        if not value:

            # If not defined,
            # the initial state of all variables is always zero

            self._initial_state = {}

            for variable in self.regulatory_variables_gen():
                variable.expression_coef = 0
                self._initial_state[variable.id] = 0

        elif not isinstance(value, dict):
            raise TypeError("The initial state must be a dict object. Use dict() to convert")

        else:

            # initial state comprehends the state of all regulatory variables in the regulatory model
            # partial setting is accepted and only the respective variables are altered

            # State or current state is an additional property to be added. Alternatively, this property keeps track
            # of the current state.

            # Either way, they both state and initial_state would change the state of all regulatory variables when
            # using the setter

            if not self._initial_state:

                self._initial_state = {}

                for variable in self.regulatory_variables_gen():

                    if variable.id in value:
                        variable.expression_coef = value[variable.id]
                        self._initial_state[variable.id] = value[variable.id]

                    else:
                        variable.expression_coef = 0
                        self._initial_state[variable.id] = 0

            else:

                for key, val in value.items():

                    if key in self.regulatory_variables:
                        self.regulatory_variables[key].expression_coef = val
                        self._initial_state[key] = val

    def initial_state_gen(self):

        """
        Get initial state
        :return: generator object of initial state items
        """

        return ((key, val) for key, val in self.initial_state.items())

    def metabolic_regulatory_genes_gen(self):

        """
        Get metabolic regulatory genes
        :return: generator object of metabolic regulatory genes values
        """

        return (val for val in self.metabolic_regulatory_genes.values())

    def metabolic_regulatory_reactions_gen(self):

        """
        Get metabolic regulatory genes
        :return: generator object of metabolic regulatory reactions values
        """

        return (val for val in self.metabolic_regulatory_reactions.values())

    def metabolic_regulatory_metabolites_gen(self):

        """
        Get metabolic regulatory genes
        :return: generator object of metabolic regulatory metabolites values
        """

        return (val for val in self.metabolic_regulatory_metabolites.values())

    def regulatory_conditions_gen(self):

        """
        Get regulatory conditions
        :return: generator object of regulatory conditions values
        """

        return (val for val in self.regulatory_conditions.values())

    def add_regulator(self, identifier,
                      name=None,
                      expression_coef=None,
                      aliases=None,
                      regulatory_interactions=None,
                      targets=None,
                      regulators=None,
                      symbol=None):

        """


        :param identifier:
        :param name:
        :param expression_coef:
        :param aliases:
        :param regulatory_interactions:
        :param targets:
        :param regulators:
        :param symbol:
        :return:
        """

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        if identifier in self.regulators:

            regulator = self.regulators[identifier]

            if name is not None:
                regulator.name = name
            if expression_coef is not None:
                regulator.expression_coef = expression_coef
            if aliases is not None:
                regulator.aliases = aliases
            if regulatory_interactions is not None:
                regulator.regulatory_interactions.update(regulatory_interactions)
            if targets is not None:
                regulator.targets.update(targets)
            if regulators is not None:
                regulator.regulators.update(regulators)
            if symbol is not None:
                regulator.symbol = symbol

            return regulator

        if identifier in self.targets:

            regulator = self.targets[identifier]

            if name is not None:
                regulator.name = name
            if expression_coef is not None:
                regulator.expression_coef = expression_coef
            if aliases is not None:
                regulator.aliases = aliases
            if regulatory_interactions is not None:
                regulator.regulatory_interactions.update(regulatory_interactions)
            if targets is not None:
                regulator.targets.update(targets)
            if regulators is not None:
                regulator.regulators.update(regulators)
            if symbol is not None:
                regulator.symbol = symbol
            regulator.is_regulator = True

            return regulator

        regulator = IntegratedRegulatoryVariable(
            identifier=identifier,
            name=name,
            expression_coef=expression_coef,
            aliases=aliases,
            regulatory_interactions=regulatory_interactions,
            targets=targets,
            regulators=regulators,
            model=self,
            cbm_model=self.cbm_simulation_interface,
            symbol=symbol,
            is_regulator=True,
            is_target=False)

        self.regulatory_variables[regulator.id] = regulator

        return regulator

    def add_target(self, identifier,
                   name=None,
                   expression_coef=None,
                   aliases=None,
                   regulatory_interactions=None,
                   targets=None,
                   regulators=None,
                   symbol=None):

        """

        :param identifier:
        :param name:
        :param expression_coef:
        :param aliases:
        :param regulatory_interactions:
        :param targets:
        :param regulators:
        :param symbol:
        :return:
        """
        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        if identifier in self.targets:

            target = self.targets[identifier]

            if name is not None:
                target.name = name
            if expression_coef is not None:
                target.expression_coef = expression_coef
            if aliases is not None:
                target.aliases = aliases
            if regulatory_interactions is not None:
                target.regulatory_interactions.update(regulatory_interactions)
            if targets is not None:
                target.targets.update(targets)
            if regulators is not None:
                target.regulators.update(regulators)
            if symbol is not None:
                target.symbol = symbol

            return target

        if identifier in self.regulators:

            target = self.regulators[identifier]

            if name is not None:
                target.name = name
            if expression_coef is not None:
                target.expression_coef = expression_coef
            if aliases is not None:
                target.aliases = aliases
            if regulatory_interactions is not None:
                target.regulatory_interactions.update(regulatory_interactions)
            if targets is not None:
                target.targets.update(targets)
            if regulators is not None:
                target.regulators.update(regulators)

            target.is_target = True

            return target

        target = IntegratedRegulatoryVariable(
            identifier=identifier,
            name=name,
            expression_coef=expression_coef,
            aliases=aliases,
            regulatory_interactions=regulatory_interactions,
            targets=targets,
            regulators=regulators,
            model=self,
            cbm_model=self.cbm_simulation_interface,
            symbol=symbol,
            is_regulator=False,
            is_target=True)

        self.regulatory_variables[target.id] = target

        return target

    def add_regulatory_interaction(self, identifier,
                                   rule,
                                   target,
                                   target_aliases=None,
                                   regulators=None):

        """

        :param identifier:
        :param rule:
        :param target:
        :param target_aliases:
        :param regulators:
        :return:
        """

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        if not isinstance(rule, str):
            raise TypeError("The regulatory rule must be a string. Use str() to convert")

        if not isinstance(target, str) and not isinstance(target, RegulatoryVariable):
            raise TypeError("Target must be a string or a RegulatoryVariable object")

        if not target_aliases:
            target_aliases = []

        if identifier in self.regulatory_interactions:
            def add_reg_inter_warn(message):
                warn(message, UserWarning, stacklevel=2)

            add_reg_inter_warn("The ID already exists in the dictionary of regulatory interactions of the regulatory "
                               "model. Ignoring the alteration")
            return

        return IntegratedRegulatoryInteraction.from_str(identifier=identifier,
                                                        rule=rule,
                                                        target=target,
                                                        target_aliases=target_aliases,
                                                        regulators=regulators,
                                                        model=self,
                                                        cbm_model=self.cbm_simulation_interface)

    def __build_gprs__(self):

        self._gprs = {}

        # A GPR object can also be created and stored under the gprs variable. This object can be much likely
        # designed as the regulatory interaction one
        # So far, the gprs_evaluater contains a list of sympy symbols as arguments, a sympy expression and sympy
        # lambda function for a given gpr rule
        self._gprs_evaluator = {}

        for reac_id in self.reactions:

            gpr = self._cbm_simulation_interface.get_gpr(reac_id)

            if not gpr:
                gpr = ''

            self._gprs[reac_id] = gpr

            self._gprs_evaluator[reac_id] = boolean_rule_from_str(gpr)

    def __infer_metabolic_regulatory_variables__(self):

        # stores all gene-like regulatory variables available in the cbm and regulatory model
        self._metabolic_regulatory_genes = {}

        # stores all reaction-like  regulatory variables available in the cbm and regulatory model
        self._metabolic_regulatory_reactions = {}

        # stores all metabolite-like  regulatory variables available in the cbm and regulatory model
        self._metabolic_regulatory_metabolites = {}

        # stores all regulatory variables (regulators only) not available in the cbm model but available in the
        # regulatory one
        self._regulatory_conditions = {}

        # map between all regulatory and metabolic entities in both models, as they might be linked by alias
        self._aliases_map = {}

        for variable in self.regulatory_variables_gen():

            if variable.cbm_model_id in self.genes:
                self._metabolic_regulatory_genes[variable.id] = variable
                self._aliases_map[variable.cbm_model_id] = variable.id

            elif variable.cbm_model_id in self.reactions:
                self._metabolic_regulatory_reactions[variable.id] = variable
                self._aliases_map[variable.cbm_model_id] = variable.id

            elif variable.cbm_model_id in self.metabolites:
                self._metabolic_regulatory_metabolites[variable.id] = variable
                self._aliases_map[variable.cbm_model_id] = variable.id

            else:
                if variable.id not in self.targets:
                    self._regulatory_conditions[variable.id] = variable

    def decode_metabolic_state(self, state):

        """
        Method responsible for decoding the RFBA metabolic state, namely the state of all metabolic genes associated
        at least with one reaction in the GPRs rule.

        :param state: dict, key is the id of the regulatory variable (metabolic target) while value can be 0, 1 or float (reactions and metabolites predicates)
        :return: SimulationResult object.
        """

        # This function decodes a given metabolic state (state of the metabolic genes) and simulates the cbm model
        # with the resulting constraints

        constraints = {}

        # for every reaction that has a valid gpr rule
        for rxn, (_, _, tree, variables, _, conditionals) in self._gprs_evaluator.items():

            if tree.value is None:
                continue

            # creates a state map (dict) for that specific rule
            state_map = {}
            true_list = []

            for arg in variables:
                if arg in self._aliases_map:
                    if self._aliases_map[arg] in state:
                        state_map[arg] = state[self._aliases_map[arg]]
                        continue
                state_map[arg] = 1

            new_state_map = {}

            for key, value in state_map.items():

                if key not in conditionals:

                    if value:
                        true_list.append(key)

                else:
                    new_state_map[key] = value

            # solving the rule
            evaluator = BooleanEvaluator(true_list, new_state_map)
            res = tree.evaluate(evaluator.f_operand, evaluator.f_operator)
            res = solution_decode(res)

            # if the reaction is off
            if not bool(res):
                constraints[rxn] = (0.0, 0.0)

        return constraints

    def simulate_cbm_model(self, objective=None, maximize=True, method=None, constraints=None):

        """

        Standard CBM model simulation in mewpy. See simulate method of the Simulation object for further detail

        :param objective: dict
        :param maximize: bool
        :param method: SimulationMethod object
        :param constraints: dict, additional constraints focused on the reactions
        :return: SimulationResult object, it contains the result of the simulate method.
        """

        self.objective = objective
        self.maximize = maximize
        self.cbm_simulation_method = method

        # broader try, as sometimes an infeasible exception is thrown by mewpy simulate interface. Sepcial when the
        # initial state of the model is fulfilled by zeros
        try:
            return self.cbm_simulation_interface.simulate(objective=self.objective, method=self.cbm_simulation_method,
                                                          maximize=self.maximize, constraints=constraints)

        except:
            return SimulationResult(self.cbm_model,
                                    0.0,
                                    {},
                                    status=SStatus.INFEASIBLE)

    def simulate(self):

        raise NotImplementedError("IntegratedModel is a base classe without clear instructions over how regulatory "
                                  "constraints should be used for simulation. Instantiate RFBA or SRFBA instead")

    def __populate_solution__(self):

        for regvar in self.regulatory_variables_gen():
            regvar._solution = regvar.expression_coef
            regvar.expression_coef = self.initial_state[regvar.id]

    def update_aliases_from_tabular_format_file(self, aliases_file,
                                                sep=None,
                                                id_col=None,
                                                aliases_cols=None,
                                                header=None):
        """

        Updating the aliases of the regulatory variables in the integrated model allows to link regulatory variables
        to metabolic entities (e.g. genes, reactions or metabolites) present in the regulatory rules.

        Linking metabolic and regulatory entities is crucial for the integrated analysis to work properly.

        :param aliases_file: str, path to aliases file in the tabular format (txt, csv, ...)
        :param sep: str, separator
        :param id_col: int, index of the regulatory variables identifiers
        :param aliases_cols: int, index of the regulatory variables aliases
        :param header: int or None, index of the header row or None if there is none
        :return:
        """

        if not isinstance(aliases_file, str):
            raise TypeError("File or file path must be a string")

        df = read_tabular_aliases(aliases_file, sep=sep, id_col=id_col, aliases_cols=aliases_cols, header=header)

        aliases_cols = [col for col in df.columns if col != 'ids']

        for row in df.index:

            new_aliases = list(df.loc[row, aliases_cols])

            if row in self.regulatory_variables:
                self.regulatory_variables[row].aliases = new_aliases
                self.regulatory_variables[row].cbm_model = None
                self.regulatory_variables[row].cbm_model = self.cbm_simulation_interface

        self.__infer_metabolic_regulatory_variables__()

    # def essential_regulatory_variables(self):
    #
    #     # TODO: Implement a method to find all essential regulatory variables following the approach essential_genes
    #
    #     pass



"""
##################################################################

GECKO E. coli model optimization

##################################################################
"""
from mewpy.model.gecko import GeckoModel
from mewpy.simulation.reframed import GeckoSimulation
from mewpy.simulation import SimulationMethod
from mewpy.simulation.simulation import SimulationResult
from mewpy.problems.gecko import GeckoKOProblem, GeckoOUProblem
from mewpy.optimization.evaluation import BPCY, WYIELD, TargetFlux
from mewpy.optimization import EA, set_default_engine
import mewpy.utils.utilities as utl
from collections import OrderedDict
from time import time


ITERATIONS = 100




def ec_gecko_ko(compound, display=False, filename=None):
    """ GECKO enzyme deletion example. 
    It runs a multi objective optimization for the increased production of a certain compound on E. Coli.
    The GECKO model is the yeast model companion from the GECKO paper "Improving the phenotype predictions
    of a yeast genome‐scale metabolic model by incorporating enzymatic constraints" https://doi.org/10.15252/msb.20167411. 
    Runs over the MEWpy implementation.
    
    :param compound: A target reaction identifier.
    :param display: Prints the best solution.
    :param filename: If given, saves the results as csv to filename.
         
    """

    import os
    dir_path = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(dir_path, '../../../examples/models/gecko')
    DATA_FILE = os.path.join(PATH, 'eciML1515_batch.xml')
    from reframed.io.sbml import load_cbmodel
    m = load_cbmodel(DATA_FILE)
    
    model = GeckoModel(m, biomass_reaction_id='R_BIOMASS_Ec_iML1515_core_75p37M',protein_pool_exchange_id='R_prot_pool_exchange',reaction_prefix='R_')
    model.set_objective({'R_BIOMASS_Ec_iML1515_core_75p37M': 1.0})
    
    
    envcond = OrderedDict()

   
    # the evaluation (objective) functions
    
    evaluator_1 = BPCY("R_BIOMASS_Ec_iML1515_core_75p37M", compound,
                       method=SimulationMethod.lMOMA)
    evaluator_2 = WYIELD("R_BIOMASS_Ec_iML1515_core_75p37M", compound)
    # The optimization problem
    problem = GeckoKOProblem(model,
                              fevaluation=[evaluator_1, evaluator_2],
                              envcond=envcond,
                              prot_prefix='R_draw_prot_',
                              candidate_max_size=6)
    
    
    # A new instance of the EA optimizer
    ea = EA(problem, max_generations=ITERATIONS)
    # runs the optimization
    final_pop = ea.run()
    # optimization results
    if display:
        individual = max(final_pop)
        best = list(problem.decode(individual.candidate).keys())
        print('Best Solution: \n{0}'.format(str(best)))
    # save final population to file
    if filename:
        print("Simplifying and saving solutions to file")
        utl.population_to_csv(problem, final_pop, filename, simplify=False)
    

def ec_gecko_ou(compound, display=False, filename=None):
    """ GECKO enzyme over/under expression example. 
    It runs a multi objective optimization for the increased production of a certain compound on E. Coli.
    The GECKO model is the yeast model companion from the GECKO paper "Improving the phenotype predictions
    of a yeast genome‐scale metabolic model by incorporating enzymatic constraints" https://doi.org/10.15252/msb.20167411. 
    Runs over the MEWpy implementation.
    
    :param compound: A target reaction identifier.
    :param display: Prints the best solution.
    :param filename: If given, saves the results as csv to filename.
         
    """
    # define the default solver
    #from reframed.solvers import set_default_solver
    #set_default_solver('gurobi')

    import os
    dir_path = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(dir_path, '../../../examples/models/gecko/')
    DATA_FILE = os.path.join(PATH, 'eciML1515_batch.xml')
    from reframed.io.sbml import load_cbmodel
    m = load_cbmodel(DATA_FILE)
    model = GeckoModel(m, biomass_reaction_id='R_BIOMASS_Ec_iML1515_core_75p37M',protein_pool_exchange_id='R_prot_pool_exchange',reaction_prefix='R_')
    model.set_objective({'R_BIOMASS_Ec_iML1515_core_75p37M': 1.0})
    
    # change protein pool bound (suggested by Leslie)
    model.reactions['R_prot_pool_exchange'].ub = 0.26 
    # define environmental consitions
    envcond = OrderedDict()

    # the evaluation (objective) functions
    evaluator_1 = BPCY("R_BIOMASS_Ec_iML1515_core_75p37M", compound,
                       method=SimulationMethod.MOMA)
    # FVA MAX is strangely very high... changing the default alpha (0.3) to compensate..
    evaluator_2 = WYIELD("R_BIOMASS_Ec_iML1515_core_75p37M", compound, alpha= 0.01 )

    evaluator_3 = TargetFlux(compound)
    
    # The optimization problem
    problem = GeckoOUProblem(model,
                              fevaluation=[evaluator_1, evaluator_2,evaluator_3],
                              envcond=envcond,
                              prot_prefix='R_draw_prot_',
                              candidate_max_size=30)
    
    # A new instance of the EA optimizer
    ea = EA(problem, max_generations=ITERATIONS, mp=True)
    # runs the optimization
    final_pop = ea.run()
    # optimization results
    if display:
        individual = max(final_pop)
        best = list(problem.decode(individual.candidate).keys())
        print('Best Solution: \n{0}'.format(str(best)))
    # save final population to file
    if filename:
        print("Simplifying and saving solutions to file")
        utl.population_to_csv(problem, final_pop, filename, simplify=False)





if __name__ == '__main__':
    """
    N_EXP = 1

    compounds = {'SUC': 'r_2056',
                 'TYR': 'r_1913',
                 'PHE': 'r_1903',
                 'TRY': 'r_1912'}

    for k, v in compounds.items():
        for _ in range(N_EXP):
            millis = int(round(time() * 1000))
            gecko_ou(v, display=False,
                     filename="gecko_{}_OU_{}.csv".format(k, millis))

    for k, v in compounds.items():
        for _ in range(N_EXP):
            millis = int(round(time() * 1000))
            gecko_ko(v, display=False,
                     filename="gecko_{}_KO_{}.csv".format(k, millis))
    """
    #from mewpy.utils.constants import EAConstants
    #EAConstants.DEBUG = True
    from reframed.solvers import set_default_solver
    set_default_solver('gurobi')
    ec_gecko_ou('R_EX_tyr__L_e',filename='gecko_ec_tyr.csv')



class RegulatoryVariable:

    def __init__(self, id=None, name=None, aliases=None):

        self._id = id
        self._name = name
        self._aliases = aliases

    def __str__(self):
        return self._id

    def __repr__(self):

        return '{}({}, {}, {}) at {}'.format(self.__class__.__name__, self._id, self._name, self._aliases, id(self))

    @property
    def id(self):
        return getattr(self, '_id', None)

    @property
    def name(self):
        return getattr(self, '_name', None)

    @property
    def aliases(self):
        return getattr(self, '_aliases', None)

    @id.setter
    def id(self, value):

        if value == self._id:
            pass

        elif not isinstance(value, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        else:
            self._id = value

    @name.setter
    def name(self, value):

        if value == self._name:
            pass

        elif not isinstance(value, str):
            raise TypeError("The name must be a string. Use str() to convert")

        else:
            self._name = value

    @aliases.setter
    def aliases(self, value):

        if value == self._aliases:
            pass

        elif not isinstance(value, list):
            raise TypeError("The aliases must be a list. Use list() to convert")

        else:
            self._aliases = value



from inspyred.ec.emo import Pareto


class IntTuppleBounder(object):
    """
    A bounder for (int,int,...) representations

    :param lower_bound: The integers lower bound.
    :param upper_bound: The integers upper bound.

    """

    def __init__(self, lower_bound, upper_bound):
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.range = [self.upper_bound[i]-self.lower_bound[i] +
                      1 for i in range(len(self.lower_bound))]

    def __call__(self, candidate, args):
        bounded_candidate = set()
        for c in candidate:
            al = []
            for i in range(len(c)):
                v = c[i] % self.range[i] + self.lower_bound[i]
                al.append(v)
            bounded_candidate.add(tuple(al))
        return bounded_candidate


class InspyredProblem:
    """Inspyred EA builder helper.

        :param problem: the optimization problem.
    """

    def __init__(self, problem):
        self.problem = problem

    def evaluate(self, solution):
        """Evaluates a single solution

            :param solution: The individual to be evaluated.
            :returns: A list with a fitness value or a Pareto object.

        """
        p = self.problem.evaluate_solution(solution)
        # single objective
        if self.problem.number_of_objectives == 1:
            return p[0]
        # multi objective
        else:
            return Pareto(p)

    def evaluator(self, candidates, args):
        """
        Evaluator
        Note: shoudn't be dependent on args to ease multiprocessing

        :param candidates: A list of candidate solutions.
        :returns: A list of Pareto fitness values or a list of fitness values.

        """
        fitness = []
        for candidate in candidates:
            p = self.evaluate(candidate)
            fitness.append(p)
        return fitness



"""
CB model optimization test Set of using inspyred EA module
"""
from mewpy.simulation import SimulationMethod, get_simulator
from mewpy.optimization.evaluation import WYIELD, BPCY
from mewpy.optimization import EA, set_default_engine
import mewpy.utils.utilities as utl
from collections import OrderedDict
from reframed.io.sbml import load_cbmodel
from time import time
import os


ITERATIONS = 300
set_default_engine('jmetal')


def load_ec():
    # E. Coli
    DIR = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(DIR, '../../../examples/models/ec/')
    DATA_FILE = os.path.join(PATH, "iJO1366SL.xml")
    NON_TARGET_FILE = os.path.join(
        PATH, "nontargets#RK#iJO1366SL#[lim-aerobic#glucose].txt")
    BIOMASS_ID = 'R_Ec_biomass_iJO1366_core_53p95M'
    O2 = 'R_EX_o2_LPAREN_e_RPAREN_'
    GLC = 'R_EX_glc_LPAREN_e_RPAREN_'

    model = load_cbmodel(DATA_FILE, flavor='cobra')

    old_obj = model.get_objective().keys()
    obj = {key: 0 for key in old_obj if key != BIOMASS_ID}
    obj[BIOMASS_ID] = 1
    model.set_objective(obj)

    non_target = [O2, GLC, 'R_ATPM']
    with open(NON_TARGET_FILE) as f:
        line = f.readline()
        while line:
            non_target.append(line.strip())
            line = f.readline()

    envcond = OrderedDict()
    envcond.update({GLC: (-10.0, 100000.0), O2: (-9.66, 100000.0)})

    simulation = get_simulator(model, envcond=envcond)
    res = simulation.simulate(method=SimulationMethod.pFBA)
    reference = res.fluxes

    return {'model': model, 'biomass': BIOMASS_ID, 'envcond': envcond, 'reference': reference, 'non_target': non_target}


def load_ec2():
    # E. Coli
    DIR = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(DIR, '../../../examples/models/ec/')
    DATA_FILE = os.path.join(PATH, "iML1515.xml")
    NON_TARGET_FILE = os.path.join(
        PATH, "nontargets#RK#iJO1366SL#[lim-aerobic#glucose].txt")
    BIOMASS_ID = 'R_BIOMASS_Ec_iML1515_core_75p37M'
    O2 = 'R_EX_o2_e'
    GLC = 'R_EX_glc__D_e'

    model = load_cbmodel(DATA_FILE, flavor='cobra')

    old_obj = model.get_objective().keys()
    obj = {key: 0 for key in old_obj if key != BIOMASS_ID}
    obj[BIOMASS_ID] = 1
    model.set_objective(obj)

    non_target = [O2, GLC, 'R_ATPM']
    with open(NON_TARGET_FILE) as f:
        line = f.readline()
        while line:
            non_target.append(line.strip())
            line = f.readline()

    envcond = OrderedDict()
    envcond.update({GLC: (-10.0, 100000.0), O2: (-9.66, 100000.0)})

    simulation = get_simulator(model, envcond=envcond)
    res = simulation.simulate(method=SimulationMethod.pFBA)
    reference = res.fluxes

    res = simulation.simulate()
    print(res)

    return {'model': model, 'biomass': BIOMASS_ID, 'envcond': envcond, 'reference': reference, 'non_target': non_target}


def load_yeast():
    # Yeast
    DIR = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(DIR, '../../../examples/models/yeast/')
    DATA_FILE = os.path.join(PATH, "iMM904SL_v6.xml")
    BIOMASS_ID = 'R_biomass_SC5_notrace'
    O2 = 'R_EX_o2_e_'
    GLC = 'R_EX_glc_e_'

    model = load_cbmodel(DATA_FILE, flavor='cobra')

    old_obj = model.get_objective().keys()
    obj = {key: 0 for key in old_obj if key != BIOMASS_ID}
    obj[BIOMASS_ID] = 1
    model.set_objective(obj)

    envcond = OrderedDict()
    envcond.update({GLC: (-10.0, 999999.0), O2: (-12.25, 100000.0)})

    simulation = get_simulator(model, envcond=envcond)
    res = simulation.simulate(method=SimulationMethod.pFBA)
    reference = res.fluxes

    return {'model': model, 'biomass': BIOMASS_ID, 'envcond': envcond, 'reference': reference, 'non_target': []}


def cb_ou(product, chassis='ec', display=False, filename=None):
    "CBModel Reaction KO SO example"
    if chassis == 'ec':
        conf = load_ec()
    elif chassis == 'ys':
        conf = load_yeast()
    else:
        raise ValueError

    BIOMASS_ID = conf['biomass']
    PRODUCT_ID = product
    model = conf['model']
    envcond = conf['envcond']
    reference = conf['reference']

    evaluator_1 = BPCY(BIOMASS_ID, PRODUCT_ID, method=SimulationMethod.lMOMA)
    evaluator_2 = WYIELD(BIOMASS_ID, PRODUCT_ID)
    print(model.genes)

    from mewpy.problems import GOUProblem
    problem = GOUProblem(model, fevaluation=[
                         evaluator_1, evaluator_2], envcond=envcond, reference=reference,
                         candidate_min_size=4, candidate_max_size=6,
                         operators=("lambda x,y: min(x,y)", "lambda x,y: max(x,y)"))

    ea = EA(problem, max_generations=ITERATIONS, visualizer=False)
    final_pop = ea.run()

    if display:
        individual = max(final_pop)
        best = list(problem.decode(individual.candidate).keys())
        print('Best Solution: \n{0}'.format(str(best)))

    if filename:
        print("Simplifying and saving solutions to file")
        utl.population_to_csv(problem, final_pop, filename, simplify=False)


def cb_ko(product, chassis='ec', display=False, filename=None):
    "CBModel Reaction KO SO example"
    if chassis == 'ec':
        conf = load_ec()
    elif chassis == 'ys':
        conf = load_yeast()
    else:
        raise ValueError

    BIOMASS_ID = conf['biomass']
    PRODUCT_ID = product
    model = conf['model']
    non_target = conf['non_target']
    envcond = conf['envcond']
    reference = conf['reference']

    evaluator_1 = BPCY(BIOMASS_ID, PRODUCT_ID, method=SimulationMethod.lMOMA)
    evaluator_2 = WYIELD(BIOMASS_ID, PRODUCT_ID)
    from mewpy.problems.genes import GKOProblem
    problem = GKOProblem(model, fevaluation=[
                         evaluator_1, evaluator_2], non_target=non_target, envcond=envcond, reference=reference)

    ea = EA(problem, max_generations=ITERATIONS, mp=True)
    final_pop = ea.run()

    if display:
        individual = max(final_pop)
        best = list(problem.decode(individual.candidate).keys())
        print('Best Solution: \n{0}'.format(str(best)))

    if filename:
        print("Simplifying and saving solutions to file")
        utl.population_to_csv(problem, final_pop, filename, simplify=False)


if __name__ == '__main__':
    RUNS = 10
    compounds_EC = {"PHE": "R_EX_phe_DASH_L_LPAREN_e_RPAREN_",
                    "TYR": "R_EX_tyr_DASH_L_LPAREN_e_RPAREN_",
                    "TRP": "R_EX_trp_DASH_L_LPAREN_e_RPAREN_"}

    #compounds_EC = {"TYR": "R_EX_tyr__L_e"}
    compounds_YS = {# "PHE": "R_EX_phe_L_e_",
                    "TYR": "R_EX_tyr_L_e_"
                    # "TRY": "R_EX_trp_L_e_"
                   }

    """
    for k, v in compounds_EC.items():
        for i in range(RUNS):
            millis = int(round(time() * 1000))
            cb_ko(v, filename="CBMODEL_{}_KO_{}.csv".format(k, millis))
    """
    for k, v in compounds_EC.items():
        for i in range(RUNS):
            millis = int(round(time() * 1000))
            cb_ou(v, filename="CBMODEL_{}_OU_{}.csv".format(k, millis))
    """
    for k, v in compounds_YS.items():
        for i in range(RUNS):
            millis = int(round(time() * 1000))
            cb_ko(v, chassis='ys', filename="CBMODEL_{}_KO_{}.csv".format(k, millis))
    """

    # from reframed.solvers import set_default_solver
    # set_default_solver('gurobi')
    """
    EAConstants.DEBUG = True
    for k, v in compounds_EC.items():
        for i in range(RUNS):
            millis = int(round(time() * 1000))
            cb_ou(v, chassis='ec', filename="CBMODEL_{}_OU_{}_.csv".format(k, millis))
    """



from jmetal.algorithm.singleobjective import GeneticAlgorithm, SimulatedAnnealing
from jmetal.algorithm.multiobjective import NSGAII, SPEA2
from jmetal.algorithm.multiobjective.nsgaiii import NSGAIII
from jmetal.algorithm.multiobjective.nsgaiii import UniformReferenceDirectionFactory
from jmetal.util.termination_criterion import StoppingByEvaluations
from jmetal.operator import BinaryTournamentSelection
from mewpy.utils.process import MultiProcessorEvaluator
from mewpy.optimization.ea import AbstractEA, Solution 
from .problem import JMetalKOProblem, JMetalOUProblem
from .observers import PrintObjectivesStatObserver, VisualizerObserver
from .operators import (ShrinkMutation,GrowMutationKO,GrowMutationOU,UniformCrossoverKO,
        UniformCrossoverOU,SingleMutationKO,SingleMutationOU,SingleMutationOULevel,MutationContainer)
from mewpy.utils.constants import EAConstants
from mewpy.utils.process import cpu_count
from random import Random
from time import time


# SOEA alternatives
soea_map ={
    'GA': GeneticAlgorithm,
    'SA': SimulatedAnnealing
}
# MOEA alternatives
moea_map = {
    'NSGAII': NSGAII,
    'SPEA2': SPEA2,
    'NSGAIII': NSGAIII
}


class EA(AbstractEA):
    """
    EA running helper for JMetal.

    
    :param problem: The optimization problem.
    :param initial_population: (list) The EA initial population.
    :param max_generations: (int) The number of iterations of the EA (stopping criteria). 
    """

    def __init__(self, problem, initial_population=[], max_generations=EAConstants.MAX_GENERATIONS, mp=True, visualizer=False, algorithm = None):

        super(EA, self).__init__(problem, initial_population=initial_population,
                                 max_generations=max_generations, mp=mp, visualizer=visualizer)
        self.algorithm_name = algorithm
        from mewpy.problems import Strategy
        self.crossover = UniformCrossoverKO(0.8, self.problem.candidate_max_size) if self.problem.strategy == Strategy.KO else UniformCrossoverOU(
            0.5, self.problem.candidate_max_size)
        mutators = []
        if self.problem.strategy == Strategy.KO:
            self.ea_problem = JMetalKOProblem(self.problem)
            mutators.append(GrowMutationKO(
                1.0, max_size=self.problem.candidate_max_size))
            mutators.append(ShrinkMutation(
                1.0, min_size=self.problem.candidate_min_size))
            mutators.append(SingleMutationKO(1.0))
        else:
            self.ea_problem = JMetalOUProblem(self.problem)
            mutators.append(GrowMutationOU(
                1.0, max_size=self.problem.candidate_max_size))
            mutators.append(ShrinkMutation(
                1.0, min_size=self.problem.candidate_min_size))
            mutators.append(SingleMutationOU(1.0))
        self.mutation = MutationContainer(0.3, mutators=mutators)
        

    def _run_so(self):
        """ Runs a single objective EA optimization ()
        """
        
        max_evaluations = self.max_generations * 100
        
        if self.algorithm_name == 'SA':
            self.mutation.probability = 1.0
            algorithm = SimulatedAnnealing(
                problem=self.ea_problem,
                mutation=self.mutation.probability, 
                termination_criterion=StoppingByEvaluations(max_evaluations=max_evaluations)
            )
            
        else:
            algorithm = GeneticAlgorithm(
                problem=self.ea_problem,
                population_size=100,
                offspring_population_size=100,
                mutation=self.mutation,
                crossover=self.crossover,
                selection=BinaryTournamentSelection(),
                termination_criterion=StoppingByEvaluations(
                    max_evaluations=max_evaluations)
            )

        algorithm.observable.register(observer=PrintObjectivesStatObserver())
        algorithm.run()

        result = algorithm.solutions
        return result

    def _run_mo(self):
        """ Runs a multi objective EA optimization
        """
        max_evaluations = self.max_generations * 100
        ncpu = cpu_count()
        if self.algorithm_name in moea_map.keys():
            f = moea_map[self.algorithm_name]
        else:
            if self.ea_problem.number_of_objectives > 2:
                self.algorithm_name== 'NSGAIII'
            else:
                f = moea_map['SPEA2']

        print(f"Running {self.algorithm_name}")
        if self.algorithm_name  and self.algorithm_name== 'NSGAIII':
            algorithm = NSGAIII(
                population_evaluator=MultiProcessorEvaluator(self.ea_problem.evaluate,ncpu),
                problem=self.ea_problem,
                population_size=100,
                mutation=self.mutation,
                crossover=self.crossover,
                termination_criterion=StoppingByEvaluations(max_evaluations=max_evaluations),
                reference_directions=UniformReferenceDirectionFactory(self.ea_problem.number_of_objectives, n_points=99)
            )

        else:
            algorithm = f(
                population_evaluator=MultiProcessorEvaluator(self.ea_problem.evaluate,ncpu),
                problem=self.ea_problem,
                population_size=100,
                offspring_population_size=100,
                mutation=self.mutation,
                crossover=self.crossover,
                termination_criterion=StoppingByEvaluations(
                    max_evaluations=max_evaluations),
            )

        if self.visualizer:
            algorithm.observable.register(observer=VisualizerObserver())
        algorithm.observable.register(observer=PrintObjectivesStatObserver())

        algorithm.run()
        result = algorithm.solutions
        return result

    def _convertPopulation(self, population):
        p = []
        for i in range(len(population)):
            # Corrects fitness values for maximization problems
            # TODO: verify each objective individualy
            if self.problem.is_maximization:
                obj = [abs(x) for x in population[i].objectives]
            else:
                obj = [x for x in population[i].objectives]
            val = set(population[i].variables[:])
            values = self.problem.translate(val)
            const = self.problem.decode(val)
            solution = Solution(values, obj, const)
            p.append(solution)
        return p



import numpy as np
import pandas as pd
import matplotlib
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


class Plot:

    def __init__(self,
                 plot_title='Pareto Aproximation',
                 reference_front=None,
                 reference_point=None,
                 axis_labels=None):
        """
        :param plot_title: Title of the graph.
        :param axis_labels: List of axis labels.
        :param reference_point: Reference point (e.g., [0.4, 1.2]).
        :param reference_front: Reference Pareto front (if any) as solutions.
        """
        self.plot_title = plot_title
        self.axis_labels = axis_labels
        self.reference_point = reference_point
        self.reference_front = reference_front
        self.dimension = None

    @staticmethod
    def get_points(solutions):
        """ Get points for each solution of the front.

        :param solutions: List of solutions where each solution is a list of fitness values
        :return: Pandas dataframe with one column for each objective and one row for each solution.
        """
        if solutions is None:
            raise Exception('Front is none!')

        points = pd.DataFrame(solutions)
        return points, points.shape[1]

    def plot(self, front, label='', normalize=False, filename=None, format='eps'):
        """ Plot any arbitrary number of fronts in 2D, 3D or p-coords.

        :param front: Pareto front or a list of them.
        :param label: Pareto front title or a list of them.
        :param normalize: If True, normalize data (for p-coords).
        :param filename: Output filename.
        :param format: Output file format.
        """
        if not isinstance(front[0][0], float):
            front = [front]

        if not isinstance(label, list):
            label = [label]

        if len(front) != len(label):
            raise Exception('Number of fronts and labels must be the same')

        dimension = len(front[0][0])

        if dimension == 2:
            self.two_dim(front, label, filename, format)
        elif dimension == 3:
            self.three_dim(front, label, filename, format)
        else:
            self.pcoords(front, normalize, filename, format)

    def two_dim(self, fronts, labels=None, filename=None, format='eps'):
        """ Plot any arbitrary number of fronts in 2D.

        :param fronts: List of fronts (containing solutions).
        :param labels: List of fronts title (if any).
        :param filename: Output filename.
        """
        n = int(np.ceil(np.sqrt(len(fronts))))
        fig = plt.figure()
        fig.suptitle(self.plot_title, fontsize=16)

        reference = None
        if self.reference_front:
            reference, _ = self.get_points(self.reference_front)

        for i, _ in enumerate(fronts):
            points, _ = self.get_points(fronts[i])

            ax = fig.add_subplot(n, n, i + 1)
            points.plot(kind='scatter', x=0, y=1, ax=ax, s=10, color='#236FA4', alpha=1.0)

            if labels:
                ax.set_title(labels[i])

            if self.reference_front:
                reference.plot(x=0, y=1, ax=ax, color='k', legend=False)

            if self.reference_point:
                for point in self.reference_point:
                    plt.plot([point[0]], [point[1]], marker='o', markersize=5, color='r')
                    plt.axvline(x=point[0], color='r', linestyle=':')
                    plt.axhline(y=point[1], color='r', linestyle=':')

            if self.axis_labels:
                plt.xlabel(self.axis_labels[0])
                plt.ylabel(self.axis_labels[1])

        if filename:
            plt.savefig(filename + '.' + format, format=format, dpi=200)

        plt.show()
        plt.close(fig)

    def three_dim(self, fronts, labels=None, filename=None, format='eps'):
        """ Plot any arbitrary number of fronts in 3D.

        :param fronts: List of fronts (containing solutions).
        :param labels: List of fronts title (if any).
        :param filename: Output filename.
        """
        n = int(np.ceil(np.sqrt(len(fronts))))
        fig = plt.figure()
        fig.suptitle(self.plot_title, fontsize=16)

        for i, _ in enumerate(fronts):
            ax = fig.add_subplot(n, n, i + 1, projection='3d')
            ax.scatter([s.objectives[0] for s in fronts[i]],
                       [s.objectives[1] for s in fronts[i]],
                       [s.objectives[2] for s in fronts[i]])

            if labels:
                ax.set_title(labels[i])

            if self.reference_front:
                ax.scatter([s.objectives[0] for s in self.reference_front],
                           [s.objectives[1] for s in self.reference_front],
                           [s.objectives[2] for s in self.reference_front])

            if self.reference_point:
                # todo
                pass

            ax.relim()
            ax.autoscale_view(True, True, True)
            ax.view_init(elev=30.0, azim=15.0)
            ax.locator_params(nbins=4)

        if filename:
            plt.savefig(filename + '.' + format, format=format, dpi=1000)

        plt.show()
        plt.close(fig)

    def pcoords(self, fronts, normalize=False, filename=None, format='eps'):
        """ Plot any arbitrary number of fronts in parallel coordinates.

        :param fronts: List of fronts (containing solutions).
        :param filename: Output filename.
        """
        n = int(np.ceil(np.sqrt(len(fronts))))
        fig = plt.figure()
        fig.suptitle(self.plot_title, fontsize=16)

        for i, _ in enumerate(fronts):
            points, _ = self.get_points(fronts[i])

            if normalize:
                points = (points - points.min()) / (points.max() - points.min())

            ax = fig.add_subplot(n, n, i + 1)
            pd.plotting.parallel_coordinates(points, 0, ax=ax)

            ax.get_legend().remove()

            if self.axis_labels:
                ax.set_xticklabels(self.axis_labels)

        if filename:
            plt.savefig(filename + '.' + format, format=format, dpi=1000)

        plt.show()
        plt.close(fig)


class StreamingPlot:

    def __init__(self,
                 plot_title='Pareto Approximation',
                 reference_front=None,
                 reference_point=None,
                 axis_labels=None):
        """
        :param plot_title: Title of the graph.
        :param axis_labels: List of axis labels.
        :param reference_point: Reference point (e.g., [0.4, 1.2]).
        :param reference_front: Reference Pareto front (if any) as solutions.
        """
        self.plot_title = plot_title
        self.axis_labels = axis_labels

        if reference_point and not isinstance(reference_point[0], list):
            reference_point = [reference_point]

        self.reference_point = reference_point
        self.reference_front = reference_front
        self.dimension = None

        import warnings
        warnings.filterwarnings("ignore", ".*GUI is implemented.*")

        self.fig, self.ax = plt.subplots()
        self.sc = None
        self.axis = None

    def plot(self, front):
        # Get data
        points, dimension = Plot.get_points(front)

        # Create an empty figure
        self.create_layout(dimension)

        # If any reference point, plot
        if self.reference_point:
            for point in self.reference_point:
                self.scp, = self.ax.plot(*[[p] for p in point], c='r', ls='None', marker='*', markersize=3)

        # If any reference front, plot
        if self.reference_front:
            rpoints, _ = Plot.get_points(self.reference_front)
            self.scf, = self.ax.plot(*[rpoints[column].tolist() for column in rpoints.columns.values],
                                     c='k', ls='None', marker='*', markersize=1)

        # Plot data
        self.sc, = self.ax.plot(*[points[column].tolist() for column in points.columns.values],
                                ls='None', marker='o', markersize=4)

        # Show plot
        plt.show(block=False)

    def update(self, front, reference_point=None):
        if self.sc is None:
            raise Exception('Figure is none')

        points, dimension = Plot.get_points(front)

        # Replace with new points
        self.sc.set_data(points[0], points[1])

        if dimension == 3:
            self.sc.set_3d_properties(points[2])

        # If any new reference point, plot
        if reference_point:
            self.scp.set_data([p[0] for p in reference_point], [p[1] for p in reference_point])

        # Re-align the axis
        self.ax.relim()
        self.ax.autoscale_view(True, True, True)

        try:
            # self.fig.canvas.draw()
            self.fig.canvas.flush_events()
        except KeyboardInterrupt:
            pass

        pause(0.01)

    def create_layout(self, dimension):
        self.fig.canvas.set_window_title(self.plot_title)
        self.fig.suptitle(self.plot_title, fontsize=16)

        if dimension == 2:
            # Stylize axis
            self.ax.spines['top'].set_visible(False)
            self.ax.spines['right'].set_visible(False)
            self.ax.get_xaxis().tick_bottom()
            self.ax.get_yaxis().tick_left()
            if self.axis_labels:
                plt.xlabel(self.axis_labels[0])
                plt.ylabel(self.axis_labels[1])
        elif dimension == 3:
            self.ax = Axes3D(self.fig)
            self.ax.autoscale(enable=True, axis='both')
            if self.axis_labels:
                self.ax.set_xlabel(self.axis_labels[0])
                self.ax.set_ylabel(self.axis_labels[1])
                self.ax.set_zlabel(self.axis_labels[2])
        else:
            raise Exception('Dimension must be either 2 or 3')

        self.ax.set_autoscale_on(True)
        self.ax.autoscale_view(True, True, True)

        # Style options
        self.ax.grid(color='#f0f0f5', linestyle='-', linewidth=0.5, alpha=0.5)


def pause(interval):
    backend = plt.rcParams['backend']

    if backend in matplotlib.rcsetup.interactive_bk:
        figManager = matplotlib._pylab_helpers.Gcf.get_active()
        if figManager is not None:
            canvas = figManager.canvas
            if canvas.figure.stale:
                canvas.draw()
            canvas.start_event_loop(interval)
            return



from mewpy.optimization.evaluation import BPCY, WYIELD
from mewpy.optimization import EA
from mewpy.simulation import SimulationMethod
from mewpy.regulation import RFBAModel
from mewpy.regulation.optorf import OptOrfProblem
import time
import os


def optorf_imc():

    """

    From Kim, J., Reed, J.L. OptORF: Optimal metabolic and regulatory perturbations for metabolic engineering of
    microbial strains. BMC Syst Biol 4, 53 (2010). https://doi.org/10.1186/1752-0509-4-53

    In the simulations, we excluded transport reactions for acetate, carbon dioxide, formate, phosphate,
    and water from consideration as eliminating transport may be challenging. In addition, ATP synthase deletion was
    excluded from consideration since the deletion resulted in a high variability in ethanol production at the
    predicted optimal growth condition. Equivantly, the deletion of focA, focB, and atp operon were excluded from the
    OptORF simulations. The OptORF approach was applied to identify metabolic engineering strategies for
    overproduction of ethanol or higher alcohols (i.e. c j = 1 for j = desired alcohol secretion) in glucose minimal
    media. Maximum glucose uptake rate (GUR) and oxygen uptake rate (OUR) are specified in order to simulate
    anaerobic growth conditions (GUR = 18.5 mmol/gDW/hr, OUR = 0 mmol/gDW/hr) [21]. A minimal growth rate was set to
    0.1 hr-1 for all simulations.

    :return:
    """

    from reframed.io.sbml import load_cbmodel
    from mewpy.simulation.reframed import Simulation

    DIR = os.path.dirname(os.path.realpath(__file__))
    cbm_model_f = os.path.join(DIR,"../../../examples/models/regulation/iJR904_srfba.xml")
    reg_model_f = os.path.join(DIR,'../../../examples/models/regulation/imc1010_v6.csv')
    aliases_f = os.path.join(DIR,'../../../examples/models/regulation/imc1010_rfba_aliases.csv')
    # env_cond_f = "../../../examples/models/regulation/imc1010_env_cond.xlsx"

    _BIOMASS_ID = 'R_BiomassEcoli'
    _O2 = 'R_EX_o2_e'
    _GLC = 'R_EX_glc_DASH_D_e'
    _CO2 = 'R_EX_co2_e'
    _FE2 = "R_EX_fe2_e"
    _H = "R_EX_h_e"
    _H2O = "R_EX_h2o_e"
    _K = "R_EX_k_e"
    _NA1 = "R_EX_na1_e"
    _NH4 = "R_EX_nh4_e"
    _PI = "R_EX_pi_e"
    _SO4 = "R_EX_so4_e"
    _SUCC = "R_EX_succ_e"
    _ETOH = "R_EX_etoh_e"

    _GLY = "R_EX_gly_e"
    _DALA = "R_EX_ala_DASH_D_e"

    model = load_cbmodel(cbm_model_f)
    model.set_objective({_BIOMASS_ID: 1})

    envcond = {_GLC: (-18.5, 100000.0),
               # _SUCC: (0.0, 100000.0),
               # _NH4: (-10.0, 100000.0),
               _O2: (0, 100000.0),
               # _CO2: (-15.0, 100000.0),
               # _PI: (-15.0, 100000.0),
               # _SO4: (-10.0, 100000.0),
               # _H: (-10.0, 100000.0),
               # _H2O: (-55.0, 100000.0)
               }

    simulation = Simulation(model, envcond=envcond)

    rfba = RFBAModel.from_tabular_format(reg_model_f, model, simulation,
                                         sep=';', id_col=1, rule_col=4, aliases_cols=[0, 2, 3], header=0)
    rfba.update_aliases_from_tabular_format_file(aliases_f, sep=';', id_col=0, aliases_cols=[1])

    initial_state = {var: 1 for var in rfba.targets}
    initial_state.update({_BIOMASS_ID: 0.1})
    rfba.initial_state = initial_state

    _PRODUCT_ID = _ETOH

    evaluator_1 = BPCY(_BIOMASS_ID, _PRODUCT_ID, method=SimulationMethod.pFBA)
    evaluator_2 = WYIELD(_BIOMASS_ID, _PRODUCT_ID)

    problem = OptOrfProblem(model, [evaluator_1, evaluator_2], rfba, candidate_max_size=6)
    
    ea = EA(problem, max_generations=100, mp=True)
    final_pop = ea.run()

    import mewpy.utils.utilities as utl

    filename = "OPTORF{}_KO_{}.csv".format(_PRODUCT_ID, "iJR904_srfba" )
    utl.population_to_csv(problem, final_pop, filename, simplify=False)


def optorf_ec():
    
    import cobra.test
    from mewpy.simulation.cobra import Simulation

    DIR = os.path.dirname(os.path.realpath(__file__))
    reg_model_f = os.path.join(DIR,'../../../examples/models/regulation/core_TRN_v2.csv')
    aliases_f = os.path.join(DIR,'../../../examples/models/regulation/core_TRN_rfba_aliases.csv')

    _BIOMASS_ID = 'Biomass_Ecoli_core'
    _O2 = 'EX_o2_e'
    _GLC = 'EX_glc__D_e'
    _FUM = 'EX_fum_e'
    _AC = 'EX_ac_e'
    _GLU = 'EX_glu__L_e'
    _LAC = 'EX_lac__D_e'
    _SUC = 'EX_succ_e'
    
    model = cobra.test.create_test_model("textbook")
    model.objective = _BIOMASS_ID

    envcond = {_GLC: (-10.0, 100000.0)}
    
    simulation = Simulation(model, envcond=envcond)
    
    

    rfba = RFBAModel.from_tabular_format(reg_model_f, model, simulation,
                                         sep=',', id_col=1, rule_col=2, aliases_cols=[0], header=0)
    rfba.update_aliases_from_tabular_format_file(aliases_f, id_col=1, aliases_cols=[0])
    

    initial_state = {var: 1 for var in rfba.targets}
    initial_state.update({_BIOMASS_ID: 0.1})
    rfba.initial_state = initial_state

    _PRODUCT_ID = _SUC

    evaluator_1 = BPCY(_BIOMASS_ID, _PRODUCT_ID, method=SimulationMethod.pFBA)
    evaluator_2 = WYIELD(_BIOMASS_ID, _PRODUCT_ID)


    problem = OptOrfProblem(model, [evaluator_1, evaluator_2], rfba, candidate_max_size=6)
    
    print(len(problem.target_list))


    ea = EA(problem, max_generations=100, mp=True)
    final_pop = ea.run()

    import mewpy.utils.utilities as utl

    filename = "OPTORF{}_KO_{}.csv".format(_PRODUCT_ID, "ec" )
    utl.population_to_csv(problem, final_pop, filename, simplify=False)




if __name__ == '__main__':
    optorf_ec()



"""
Finds expression levels for modifications suggested in the literature
"""
from mewpy.simulation.reframed import GeckoSimulation, Simulation
from mewpy.simulation import SimulationMethod
from mewpy.optimization.evaluation import WYIELD, BPCY, TargetFlux
from mewpy.optimization import EA, set_default_engine
from mewpy.model.gecko import GeckoModel
import mewpy.utils.utilities as utl
from collections import OrderedDict
from reframed.io.sbml import load_cbmodel
from reframed.core.cbmodel import CBModel
from random import Random
from time import time
import os


# Number of iterations 
ITERATIONS = 10 



def load_ec():
    # E. Coli
    DIR = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(DIR, '../../../examples/models/ec/')
    DATA_FILE = os.path.join(PATH, "iJO1366SL.xml")
    BIOMASS_ID = 'R_Ec_biomass_iJO1366_core_53p95M'
    O2 = 'R_EX_o2_LPAREN_e_RPAREN_'
    GLC = 'R_EX_glc_LPAREN_e_RPAREN_'

    model = load_cbmodel(DATA_FILE, flavor='cobra')

    old_obj = model.get_objective().keys()
    obj = {key: 0 for key in old_obj if key != BIOMASS_ID}
    obj[BIOMASS_ID] = 1
    model.set_objective(obj)

    envcond = OrderedDict()
    envcond.update({GLC: (-10.0, 100000.0), O2: (-9.66, 100000.0)})
    return {'model': model, 'biomass': BIOMASS_ID, 'envcond': envcond}


def load_yeast():
    # Yeast
    DIR = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(DIR, '../../../examples/models/yeast/')
    DATA_FILE = os.path.join(PATH, "iMM904SL_v6.xml")
    BIOMASS_ID = 'R_biomass_SC5_notrace'
    O2 = 'R_EX_o2_e_'
    GLC = 'R_EX_glc_e_'

    model = load_cbmodel(DATA_FILE, flavor='cobra')

    old_obj = model.get_objective().keys()
    obj = {key: 0 for key in old_obj if key != BIOMASS_ID}
    obj[BIOMASS_ID] = 1
    model.set_objective(obj)

    envcond = OrderedDict()
    envcond.update({GLC: (-10.0, 999999.0), O2: (-12.25, 100000.0)})

    simulation = Simulation(model, envcond=envcond)
    res = simulation.simulate(method=SimulationMethod.pFBA)
    reference = res.fluxes

    return {'model': model, 'biomass': BIOMASS_ID, 'envcond': envcond, 'reference': reference, 'non_target': []}


def cb_ou(product, modification_targets, chassis='ec', display=False, filename=None):
    
    if chassis == 'ec':
        conf = load_ec()
    elif chassis == 'ys':
        conf = load_yeast()
    else:
        raise ValueError

    BIOMASS_ID = conf['biomass']
    PRODUCT_ID = product
    model = conf['model']
    envcond = conf['envcond']
    evaluator_1 = BPCY(BIOMASS_ID, PRODUCT_ID, method=SimulationMethod.lMOMA)
    evaluator_2 = WYIELD(BIOMASS_ID, PRODUCT_ID)
    from mewpy.problems import GOUProblem
    problem = GOUProblem(model, fevaluation=[
                         evaluator_1, evaluator_2], envcond=envcond, candidate_min_size=len(modification_targets),candidate_max_size=len(modification_targets), target = modification_targets)

    ea = EA(problem, max_generations=ITERATIONS, visualizer=False)
    final_pop = ea.run()

    if display:
        individual = max(final_pop)
        best = list(problem.decode(individual.candidate).keys())
        print('Best Solution: \n{0}'.format(str(best)))

    if filename:
        print("Simplifying and saving solutions to file")
        utl.population_to_csv(problem, final_pop, filename, simplify=False)




if __name__ == '__main__':

    from mewpy.utils.constants import EAConstants
    # Define the number of CPUs to be used in parallel evaluations.
    EAConstants.NUM_CPUS = 10
    
    # change the EA engine and MOEA
    from mewpy.optimization import set_preferred_EA, set_default_engine
    set_preferred_EA('SPEA2')
    set_default_engine('jmetal')

    compounds_EC = {"PHE": "R_EX_phe_DASH_L_LPAREN_e_RPAREN_",
                    "TYR": "R_EX_tyr_DASH_L_LPAREN_e_RPAREN_",
                    "TRP": "R_EX_trp_DASH_L_LPAREN_e_RPAREN_"}

    compounds_YS = {"PHE": "R_EX_phe_L_e_",
                    "TYR": "R_EX_tyr_L_e_",
                    "TRY": "R_EX_trp_L_e_"
                   }
    
    # Can also be used to find expression values for
    # modifications found in the literature.
    millis = int(round(time() * 1000))
    modification_targets = ["G_YBR166C","G_YNL241C","G_YBR249C","G_YDR380W"]
    cb_ou("R_EX_tyr_L_e_", modification_targets ,chassis='ys', filename="CBMODEL_{}_OU_{}_.csv".format("Tyr", millis))



from reframed.core.cbmodel import CBModel, CBReaction
from reframed.io.sbml import load_cbmodel
from reframed.core.model import AttrOrderedDict, Metabolite
from mewpy.utils.constants import ModelConstants
from collections import OrderedDict
from six import iteritems, string_types
from itertools import chain
import pandas as pd
import numpy as np
from math import isinf
import logging
import copy
import os
import re



class ModelList(object):
    '''Auxilary class to load predifined ecYeast7 models
    '''


    def __init__(self):
        self.DATA_FILES = os.path.join(os.path.dirname(__file__), 'data')
        self.PROTEINS_FILE = os.path.join(self.DATA_FILES, 'proteins.txt')
        self.model_files = dict((re.findall(r'_(.*).xml', f)[0], f) for f in os.listdir(self.DATA_FILES) if f.endswith('.xml'))
        self.models = {}


    def __getitem__(self, item):
        """Get a bundled GECKO model.
        
        :param item: basestring. Either 'single-pool' for the single-protein pool ecYeastGEM model or 'multi-pool' for individually modeled protein pools.

        """
        try:
            file_name = self.model_files[item]
        except KeyError:
            raise KeyError('model name must be one of {}'.format(', '.join(list(self.model_files))))
        if file_name not in self.models:
            model = load_cbmodel(os.path.join(os.path.dirname(__file__), 'data/{}'.format(file_name)))
            self.simplify_model(model)
            self.models[file_name] = model
        return self.models[file_name]


    def simplify_model(self,model):
        met_copy = AttrOrderedDict()
        for key, val in model.metabolites.items():
            k = key.replace('__91__', '_').replace('__93__', '') 
            val.id = k
            met_copy[k] = val
        model.metabolites = met_copy
        for rxn in model.reactions.keys():
            stoi = model.reactions[rxn].stoichiometry
            nstoi = OrderedDict()
            for key , v in stoi.items():
                k = key.replace('__91__', '_').replace('__93__', '')
                nstoi[k] = v
            model.reactions[rxn].stoichiometry = nstoi
        for rxn in model.reactions.values():
            if isinf(rxn.ub):
                rxn.ub = ModelConstants.REACTION_UPPER_BOUND


    def protein_properties(self):
        return pd.read_csv(self.PROTEINS_FILE, index_col=0)


class GeckoModel(CBModel):
    """Class for representing GECKO models.  Addapted from the original (https://github.com/SysBioChalmers/GECKO/) to reframed

    Implement a model class for Genome-scale model to account for Enzyme Constraints, using Kinetics and Omics [1]_.


    
    :param model: str, A CBModel to apply protein constraints to. Can be 'single-pool' for the bundled ecYeast7 model using only a single pool for all proteins, or 'multi-pool' for the model that has separate pools for all measured proteins. 
    :param protein_properties: pd.DataFrame, A data frame that defined molecular weight (g/mol) 'mw', for 'uniprot' proteins and their average 'abundance' in ppm.
    :param sigma: float, The parameter adjusting how much of a protein pool can take part in reactions. Fitted parameter, default is optimized for chemostat experiment in [1]_.
    :param gam: float, The growth associated maintenance cost in mmol / gDW. Default fitted for yeast 8.1.3.
    :param amino_acid_polymerization_cost: float, The cost for turning amino-acids in proteins in mmol / g. Default taken from [2]_.
    :param carbohydrate_polymerization_cost: float, The cost for turning monosaccharides in polysaccharides in mmol / g. Default taken from [2]_.
    :param c_base: float, The carbohydrate content at dilution rate 0.1 / h. Default taken from yeast 8.1.3.
    :param biomass_reaction_id: str, The identifier for the biomass reaction.
    :param protein_reaction_id: str, The identifier for the protein reaction. 
    :param carbohydrate_reaction_id: str, The identifier for the carbohydrate reaction.
    :param protein_pool_exchange_id: str, The identifier of the protein pool exchange reaction.
    :param common_protein_pool_id: str, The identifier of the metabolite representing the common protein pool.


    References:
    
    .. [1] Benjamin J. Sanchez, Cheng Zhang, Avlant Nilsson, Petri-Jaan Lahtvee, Eduard J. Kerkhoven, Jens Nielsen (
       2017). Improving the phenotype predictions of a yeast genome-scale metabolic model by incorporating enzymatic
       constraints. [Molecular Systems Biology, 13(8): 935, http://www.dx.doi.org/10.15252/msb.20167411

       [2] J. Förster, I. Famili, B. Ø. Palsson and J. Nielsen, Genome Res., 2003, 244–253.

    """

    def __init__(self, model, protein_properties=None,
                 sigma=0.46, c_base=0.3855, gam=36.6, amino_acid_polymerization_cost=37.7,
                 carbohydrate_polymerization_cost=12.8, biomass_reaction_id='r_4041',
                 protein_reaction_id='r_4047', carbohydrate_reaction_id='r_4048',
                 protein_pool_exchange_id='prot_pool_exchange', common_protein_pool_id='prot_pool',
                 reaction_prefix = ''):
        
        # load predifined models
        model_list = ModelList()
        if isinstance(model, string_types):
            model = model_list.__getitem__(model)
        elif isinstance(model, CBModel):
            model_list.simplify_model(model)
        else:
            raise ValueError('Model should be a string denomination or a CBModel instance')
               
        super(GeckoModel,self).__init__(model.id)

        # import CBModel's data
        self.compartments = copy.deepcopy(model.compartments)
        self.metabolites = copy.deepcopy(model.metabolites)
        self.reactions = copy.deepcopy(model.reactions)
        self.genes = copy.deepcopy(model.genes)


        # biomass reaction id (str)
        if biomass_reaction_id not in self.reactions:
            raise RuntimeError(f"Reaction {biomass_reaction_id} is not in the model")
        self.biomass_reaction = biomass_reaction_id
        
        # protein reaction id (CBReaction) 
        try:
            self.protein_reaction = self.reactions[protein_reaction_id]
        except KeyError:
            logging.warning(f"Protein reaction {protein_reaction_id} is not in the model")
        

        # carbohydrate reaction id (CBReaction) 
        try:
            self.carbohydrate_reaction = self.reactions[carbohydrate_reaction_id]
        except KeyError:
            logging.warning(f"Carbohydrate reaction {carbohydrate_reaction_id} is not in the model")
        
        
        # protein properties DataFrame
        if protein_properties:
            self.protein_properties = protein_properties
        else:
            model_list = ModelList()
            self.protein_properties = model_list.protein_properties()
        
        
        # Metabolite of common protein pool 
        try:
            self.common_protein_pool = self.metabolites[common_protein_pool_id]
        except KeyError:
            self.common_protein_pool = Metabolite(common_protein_pool_id)
            self.metabolites[common_protein_pool_id] = self.common_protein_pool
        
        
        # Reaction identified as protein pool exchange
        if protein_pool_exchange_id in self.reactions.keys():
            self.protein_pool_exchange = self.reactions[protein_pool_exchange_id]
        else:
            self.protein_pool_exchange = CBReaction(protein_pool_exchange_id)
            self.protein_pool_exchange.stoichiometry.update({self.common_protein_pool.id: 1.})
            self.protein_pool_exchange.set_flux_bounds(0,ModelConstants.REACTION_UPPER_BOUND)
            self.add_reaction(self.protein_pool_exchange)
        
        
    
        # multi-pool
        s_protein_exchange = "^"+reaction_prefix+"prot_(.*)_exchange$"
        self.protein_exchange_re = re.compile(s_protein_exchange)
        s_pool_protein_exchange = "^"+reaction_prefix+"draw_prot_(.*)$"
        self.pool_protein_exchange_re = re.compile(s_pool_protein_exchange)
        self.concentrations = pd.Series(np.nan, index=self.proteins)
        
        # passed parameters
        self.gam = gam
        self.amino_acid_polymerization_cost = amino_acid_polymerization_cost
        self.carbohydrate_polymerization_cost = carbohydrate_polymerization_cost
        self.c_base = c_base
        self.sigma_saturation_factor = sigma

        # initialize remaining variables
        self.measured_ggdw = None
        self.fp_fraction_protein = None
        self.fc_carbohydrate_content = None
        self.p_total = None
        self.c_total = None
        self.p_base = None
        self.fn_mass_fraction_unmeasured_matched = None
        self.fs_matched_adjusted = None
        self.p_measured = None
        self.f_mass_fraction_measured_matched_to_total = None
        self.fm_mass_fraction_matched = None
        self.uniprot = {}
        # protein reverse reactions mapping
        self._protein_rev_reactions = None



    def fraction_to_ggdw(self, fraction):
        """Convert protein measurements in mass fraction of total to g protein / g DW.

        
        :param fraction: pd.Series, Data of protein measurements which are absolute quantitative fractions of the total amount of these measured proteins. Normalized to sum == 1.
        :returns: pd.Series, g protein / g DW for the measured proteins

        """
        # measurements should be quantitative fractions of the total measured proteins, normalized to unit-length
        fraction = fraction / fraction.sum()
        fraction_measured = self.protein_properties.loc[list(fraction.index),'abundance'].sum()
        p_measured = self.p_total * fraction_measured
        return fraction.apply(lambda x: x * p_measured)


    def limit_proteins(self, fractions=None, ggdw=None, p_total=0.448, p_base=0.46):
        """Apply proteomics measurements to model.

        Apply measurements in the form of fractions of total of the measured proteins, or directly as g / gDW. Must
        supply exactly one of `fractions` or `ggdw`.

        
        :param fractions: pd.Series, Protein abundances in fraction of total (normalized to sum to 1). Ignored if `ggdw` is also supplied.
        :param ggdw: pd.Series, Protein abundances in g / gDW
        :param p_total: float, measured total protein fraction in cell in g protein / g DW. Should be measured for each experiment, the default here is taken from [2]_.
        :param p_base: float, protein content at dilution rate 0.1 / h in g protein / g DW. Default taken from yeast 8.1.3.

        References:
        
        .. [2] Benjamin J. Sanchez, Cheng Zhang, Avlant Nilsson, Petri-Jaan Lahtvee, Eduard J. Kerkhoven, Jens Nielsen (
           2017). Improving the phenotype predictions of a yeast genome-scale metabolic model by incorporating enzymatic
           constraints. [Molecular Systems Biology, 13(8): 935, http://www.dx.doi.org/10.15252/msb.20167411


        """
        self.p_total = p_total
        self.p_base = p_base
        self.c_total = self.c_base + self.p_base - self.p_total
        self.fp_fraction_protein = self.p_total / self.p_base
        self.fc_carbohydrate_content = self.c_total / self.c_base
        self.measured_ggdw = self.fraction_to_ggdw(fractions) if ggdw is None else ggdw

        # * section 2.5
        # 1. define mmmol_gdw as ub for measured proteins (multi-pool)

        for protein_id, value in iteritems(self.measured_ggdw):
            try:
                mmol_gdw = value / (self.protein_properties.loc[protein_id, 'mw'] / 1000)
                rxn = self.reactions['prot_{}_exchange'.format(protein_id)]
                self.uniprot[rxn.id] = protein_id
            except KeyError:
                pass
            else:
                self.concentrations[protein_id] = value
                rxn.set_flux_bounds(0, mmol_gdw)
        
        # 2. p_measured is aggregate mass of all matched proteins
        self.p_measured = self.concentrations.sum()
        # 3. fm, mass fraction of measured proteins in the model over total
        self.fm_mass_fraction_matched = self.p_measured / self.p_total
        # 4. mass fraction of unmeasured proteins in the model over all proteins not matched to model
        self.fn_mass_fraction_unmeasured_matched = (
            self.protein_properties.loc[list(self.unmeasured_proteins)].prod(axis=1).sum() /
            self.protein_properties.prod(axis=1).sum()
        )
        self.f_mass_fraction_measured_matched_to_total = (
            self.fn_mass_fraction_unmeasured_matched / (1 - self.fm_mass_fraction_matched))
        # 5. constrain unmeasured proteins by common pool
        self.constrain_pool()
        self.adjust_biomass_composition()




    
    def constrain_pool(self):
        """Constrain the draw reactions for the unmeasured (common protein pool) proteins.

        Proteins without their own protein pool are collectively constrained by the common protein pool. Remove
        protein pools for all proteins that don't have measurements, along with corresponding draw reactions,
        and add these to the common protein pool and reaction.
        
        """
        new_reactions = []
        to_remove = []
        # * section 2.5.1
        # 1. and 2. introduce `prot_pool` and exchange reaction done in __init__
        # 3. limiting total usage with the unmeasured amount of protein
        # looks like the matlab code:
        # self.fs_matched_adjusted = ((self.p_total - self.p_measured) / self.p_base *
        #                             self.f_mass_fraction_measured_matched_to_total *
        #                             self.sigma_saturation_factor)
        # but this gives results more like reported:
        self.fs_matched_adjusted = ((self.p_total - self.p_measured) *
                                    self.f_mass_fraction_measured_matched_to_total *
                                    self.sigma_saturation_factor)
        self.protein_pool_exchange.set_flux_bounds(0, self.fs_matched_adjusted)                            

        # 4. Remove other enzyme usage reactions and replace with pool exchange reactions
        average_mmw = self.protein_properties['mw'].mean() / 1000.
        for protein_id in self.unmeasured_proteins:
            prt = 'prot_{}_exchange'.format(protein_id)
            if prt in self.reactions: 
              to_remove.append(prt)  
            draw_reaction_id = 'draw_prot_{}'.format(protein_id)
            if draw_reaction_id not in self.reactions.keys():
                draw_rxn = CBReaction(draw_reaction_id)
                # defines bounds
                draw_rxn.set_flux_bounds(0,1000)
                self.uniprot[draw_rxn.id] = protein_id
                protein_pool = self.metabolites['prot_{}_c'.format(protein_id)]
                try:
                    mmw = self.protein_properties.loc[protein_id, 'mw'] / 1000.
                except KeyError:
                    mmw = average_mmw
                metabolites = {self.common_protein_pool.id: -mmw, protein_pool.id: 1}
                draw_rxn.stoichiometry.update(metabolites)
                new_reactions.append(draw_rxn)
        for rxn in new_reactions:
            self.add_reaction(rxn)
        self.remove_reactions(to_remove)




    def adjust_biomass_composition(self):
        """Adjust the biomass composition.

        After changing the protein and carbohydrate content based on measurements, adjust the corresponding
        coefficients of the biomass reaction.
        
        """
        for met in self.protein_reaction.stoichiometry:
            is_prot = 'protein' in self.metabolites[met].name
            if not is_prot:
                coefficient = self.fp_fraction_protein * self.protein_reaction.stoichiometry[met]
                self.protein_reaction.stoichiometry[met] = coefficient

        for met in self.carbohydrate_reaction.stoichiometry:
            is_carb = 'carbohydrate' in self.metabolites[met].name
            if not is_carb:
                coefficient = self.fc_carbohydrate_content * self.carbohydrate_reaction.stoichiometry[met]
                self.carbohydrate_reaction.stoichiometry[met] = coefficient

        for met in self.reactions[self.biomass_reaction].stoichiometry:
            sign = -1 if self.reactions[self.biomass_reaction].stoichiometry[met] < 0 else 1
            is_atp = 'ATP' in self.metabolites[met].name
            is_adp = 'ADP' in self.metabolites[met].name
            is_h2o = 'H2O' in self.metabolites[met].name
            is_h = 'H+' in self.metabolites[met].name
            is_p = 'phosphate' in self.metabolites[met].name
            if is_atp or is_adp or is_h2o or is_h or is_p:
                coefficient = sign * (self.gam +
                                      self.amino_acid_polymerization_cost * self.p_total +
                                      self.carbohydrate_polymerization_cost * self.c_total)
                self.reactions[self.biomass_reaction].stoichiometry[met] = coefficient


    
    def adjust_pool_bounds(self, min_objective=0.05, inplace=False, tolerance=1e-9):
        """Adjust protein pool bounds minimally to make model feasible.

        Bounds from measurements can make the model non-viable or even infeasible. Adjust these minimally by minimizing
        the positive deviation from the measured values.
        
        :param min_objective: float, The minimum value of for the ojective for calling the model viable.
        :param inplace: bool, Apply the adjustments to the model.
        :param tolerance: float, Minimum non-zero value. Solver specific value.
        :returns: pd.DataFrame, Data frame with the series 'original' bounds and the new 'adjusted' bound, and the optimized 'addition'.

        """
        from reframed.solvers import solver_instance
        solver = solver_instance(self)
        solver.add_constraint(
            'constraint_objective', self.get_objective, sense='>', rhs=min_objective)
        for pool in self.individual_protein_exchanges:
            solver.add_variable('pool_diff_' + pool.id, lb=0)
            solver.add_variable('measured_bound_' + pool.id,
                                lb=pool.upper_bound, ub=pool.upper_bound)

        
        #with self.model as model:
        #    problem = model.problem
        #    constraint_objective = problem.Constraint(model.objective.expression, name='constraint_objective',
        #                                              lb=min_objective)
        #    to_add = [constraint_objective]
        #    new_objective = S.Zero
        #    for pool in model.individual_protein_exchanges:
        #        ub_diff = problem.Variable('pool_diff_' + pool.id, lb=0, ub=None)
        #        current_ub = problem.Variable('measured_bound_' + pool.id, lb=pool.upper_bound, ub=pool.upper_bound)
        #        constraint = problem.Constraint(pool.forward_variable - current_ub - ub_diff, ub=0,
        #                                        name='pool_ub_' + pool.id)
        #        to_add.extend([ub_diff, current_ub, constraint])
        #        new_objective += ub_diff
        #        pool.bounds = 0, 1000.
        #    model.add_cons_vars(to_add)
        #    model.objective = problem.Objective(new_objective, direction='min')
        #    model.slim_optimize(error_value=None)
        #    primal_values = model.solver.primal_values
        #adjustments = [(pool.id, primal_values['pool_diff_' + pool.id], pool.upper_bound)
        #               for pool in model.individual_protein_exchanges
        #               if primal_values['pool_diff_' + pool.id] > tolerance]
        #result = pd.DataFrame(adjustments, columns=['reaction', 'addition', 'original'])
        #result['adjusted'] = result['addition'] + result['original']
        #if inplace:
        #    for adj in result.itertuples():
        #        model.reactions.get_by_id(adj.reaction).upper_bound = adj.adjusted
        #return result
        


    @property
    def measured_proteins(self):
        """Get the identifiers of the measured proteins.

        :returns: frozenset, The identifiers for the unmeasured proteins.

        """
        return frozenset(self.concentrations[self.concentrations.notnull()].index)

    @property
    def unmeasured_proteins(self):
        """Get the identifiers of the proteins .

        :returns: frozenset, The protein identifiers for the measured proteins.

        """
        return frozenset(self.concentrations[self.concentrations.isnull()].index)

    @property
    def proteins(self):
        """Get all proteins.

        :returns: frozenset, The set of all proteins identifiers.

        """
        return self.individual_proteins.union(self.pool_proteins)

    @property
    def individual_proteins(self):
        """Get the identifiers for the proteins with their individual abundance pool.

        :returns: frozenset, The set of proteins that have a defined separate pool exchange reaction.

        """
        return frozenset(chain.from_iterable(re.findall(self.protein_exchange_re, rxn)
                                             for rxn in self.protein_exchanges))

    @property
    def pool_proteins(self):
        """Get proteins modeled by common protein pool.

        :returns: frozenset, The set of proteins that have a defined draw reaction.

        """
        return frozenset(chain.from_iterable(re.findall(self.pool_protein_exchange_re, rxn)
                                             for rxn in self.protein_exchanges))

    @property
    def individual_protein_exchanges(self):
        """Individual protein-exchange reactions.

        :returns: frozenset, Set of protein exchange reactions with individual pools

        """
        return (frozenset(rxn for rxn in self.reactions.keys()
                          if re.match(self.protein_exchange_re, rxn)) - {self.protein_pool_exchange.id})

    @property
    def pool_protein_exchanges(self):
        """Protein-exchange reactions by single pool.

        :returns: frozenset, Set of protein exchange reactions for single pool reactions.

        """
        return (frozenset(rxn for rxn in self.reactions
                          if re.match(self.pool_protein_exchange_re, rxn)) -
                {self.protein_pool_exchange})



    @property
    def protein_exchanges(self):
        """Protein-exchange reactions.
            fs_matched_adjusted
        
        :returns: frozenset, Set of protein exchange reactions (individual and common protein pool reactions).

        """
        return self.individual_protein_exchanges.union(self.pool_protein_exchanges)




    @property
    def protein_rev_reactions(self):
        """
        Pairs of reverse reactions associated with a protein
        
        :returns: dictionaty, A dictionary which identifies for each protein (key) the list of reversible reactions pairs.
            
        """
        if not self._protein_rev_reactions:
            proteins = self.proteins
            reactions = self.reactions
            in_sub = {}
            for p in proteins:
                sub = []
                for r_id, rxn in reactions.items():
                    lsub = rxn.get_substrates()
                    for m in lsub:
                        if p in m:
                            sub.append(r_id)
                in_sub[p]=sub
            pairs ={}    
            for k,s in in_sub.items():
                revs = [r for r in s if '_REV' in r]
                if len(revs)>0:
                    for r in revs:
                        l = [ a for a in s if r.replace('_REV','') == a]
                        l.append(r) 
                        if len(l)==2:
                            if k in pairs.keys():
                                pairs[k].append((l[0],l[1]))
                            else:
                                pairs[k]=[(l[0],l[1])]
            self._protein_rev_reactions = pairs
        return self._protein_rev_reactions    



                 


   










    

    


from mewpy.utils.constants import EAConstants
from .ea import Solution

engines = dict()

try:
    from .inspyred.ea import EA as InspyredEA
    engines['inspyred'] = InspyredEA
except ImportError:
    print("inspyred not available")

try:
    from .jmetal.ea import EA as JMetalEA
    engines['jmetal'] = JMetalEA
except ImportError:
    print("jmetal not available")


algorithms ={ 'inspyred' : ['SA','GA','NSGAII'],
              'jmetal'   : ['SA','GA','NSGAII','SPEA2','NSGAIII']
            }


default_engine = None
preferred_EA = 'NSGAII'

def get_default_engine():

    global default_engine

    if default_engine:
        return default_engine

    engine_order = ['inspyred', 'jmetal']

    for engine in engine_order:
        if engine in list(engines.keys()):
            default_engine = engine
            break

    if not default_engine:
        raise RuntimeError("No EA engine available.")

    return default_engine


def set_default_engine(enginename):
    """ Sets default EA engine.
   
    :param str enginename: Optimization engine (currently available: 'inspyred', 'jmetal')
    """

    global default_engine

    if enginename.lower() in list(engines.keys()):
        default_engine = enginename.lower()
    else:
        raise RuntimeError(f"EA engine {enginename} not available.")


def set_preferred_EA(algorithm):
    """Defines de preferred MOEA.
    
    :param str algorithm: The name of the preferred algorithm.
    """
    global preferred_EA
    global default_engine

    if algorithm in algorithms[get_default_engine()]:
        preferred_EA = algorithm
    else:
        for eng in engines.keys():
            if algorithm in algorithms[eng]:
                preferred_EA = algorithm            
                default_engine = eng
                return
        raise ValueError(f"Algorithm {algorithm} is unavailable.")


def get_preferred_EA():
    """
    :returns: The name of the preferred MOEA.
    """
    global preferred_EA
    return preferred_EA


def get_available_engines():
    """
    :returns: The list of available engines.
    """
    return list(engines.keys())


def get_available_algorithms():
    """
    :returns: The list of available MOEAs.
    """
    l = set()
    for engine in engines.keys():
        l.union(set(engines[engine]))
    return list(l)



def EA(problem, initial_population=[], max_generations=EAConstants.MAX_GENERATIONS, mp=True, visualizer=False, algorithm = None ):
    """
    EA running helper. Returns an instance of the EA that reflects the global user configuration settings such as preferred engine and algorithm.

    :param problem: The optimization problem.
    :param list initial_population: The EA initial population.
    :param int max_generations: The number of iterations of the EA (stopping criteria).
    :param bool mp: If multiprocessing should be used. 
    :param bool visualizer: If the pareto font should be displayed. Requires a graphic environment.
    :returns: An instance of an EA optimizer.
    
    """

    if len(engines) == 0:
        raise RuntimeError('Inspyred or JMetal packages are required.')
    engine = engines[get_default_engine()]
    
    if algorithm is None or algorithm not in get_available_algorithms():
        algorithm=get_preferred_EA()
     
    return engine(problem, initial_population=initial_population, max_generations=max_generations, mp=mp, visualizer=visualizer,algorithm=algorithm)



from mewpy.problems.problem import AbstractKOProblem


class OptOrfProblem(AbstractKOProblem):
    """
    EA OptOrf problem based on a RFBA problem.
    Currently only consideres deletions.
    """

    def __init__(self, model, fevaluation, regmodel, **kwargs):

        super(OptOrfProblem, self).__init__(model, fevaluation, **kwargs)

        self.regmodel = regmodel

        # ignore user defined target list
        self._trg_list = None

    def _build_target_list(self):

        """
        The EA target list is the combination [metabolic regulatory genes]+[regulators] for a integrated model which
        is the basis for a RFBA model
        """

        self._trg_list = self.regmodel.genes + [regulator for regulator in self.regmodel.regulators.keys()
                                                if regulator not in self.regmodel.metabolic_regulatory_reactions and
                                                regulator not in self.regmodel.metabolic_regulatory_metabolites and
                                                regulator not in self.regmodel.metabolic_regulatory_genes and
                                                regulator not in self.regmodel.regulatory_conditions]

    def decode(self, candidate):
        """

        OptOrf decode based on a RFBA model

        :param candidate:
        :return:
        """

        regulatory_candidate = {}
        metabolic_candidate = {}

        for t_idx in candidate:

            if t_idx > len(self.regmodel.genes) - 1:
                regulatory_candidate[self.target_list[t_idx]] = 0
            else:
                metabolic_candidate[self.target_list[t_idx]] = 0

        # solve first the regulatory model
        state = self.regmodel.solve_regulatory_model(regulatory_candidate, silently=True)

        # update the resulting state with the metabolic share of the candidate
        state.update(metabolic_candidate)

        # find the affected reactions
        constraints = self.regmodel.decode_metabolic_state(state)

        return constraints



"""
Simulation for COBRApy models
"""
from cobra.core.model import Model
from cobra.core.solution import Solution
from geckopy.gecko import GeckoModel
from cobra.flux_analysis import pfba, moma, room
from mewpy.simulation import SimulationMethod, SStatus
from mewpy.simulation.simulation import Simulator, SimulationResult, ModelContainer
from mewpy.utils.constants import ModelConstants
from mewpy.utils.parsing import evaluate_expression_tree
from collections import OrderedDict
import numpy as np
import logging


LOGGER = logging.getLogger(__name__)


class CobraModelContainer(ModelContainer):
    """ A basic container for COBRApy models.

    :param model: A metabolic model.

    """
    def __init__(self, model: Model):
        if not isinstance(model, Model):
            raise ValueError("The model is not an instance of cobrapy Model")
        self.model = model

    @property
    def reactions(self):
        return [rxn.id for rxn in self.model.reactions]

    @property
    def genes(self):
        return [gene.id for gene in self.model.genes]

    @property
    def metabolites(self):
        return [met.id for met in self.model.metabolites]

    @property
    def medium(self):
        return self.model.medium

    @property
    def compartments(self):
        return self.model._compartments

    def get_gpr(self, reaction_id):
        """Returns the gpr rule (str) for a given reaction ID.

        :param str reaction_id: The reaction identifier.
        :returns: A string representation of the GPR rule.

        """
        if reaction_id not in self.reactions:
            raise ValueError(f"Reactions {reaction_id} does not exist")
        reaction = self.model.reactions.get_by_id(reaction_id)
        if reaction.gene_reaction_rule:
            return str(reaction.gene_reaction_rule)
        else:
            return None

    def get_drains(self):
        rxns = [r.id for r in self.model.exchanges]
        return rxns


class Simulation(CobraModelContainer, Simulator):
    """Generic Simulation class for cobra Model.
       Defines the simulation conditions, and makes available a set of methods.

    :param model: An metabolic model instance.

    Optional:

    :param objective: The model objective.
    :param dic envcond: Dictionary of environmental conditions.
    :param dic constraints: A dictionary of reaction constraints.
    :param solver: An instance of the LP solver.
    :param dic reference: A dictionary of the wild type flux values.

    """

    def __init__(self, model: Model, objective=None, envcond=None, constraints=None,  solver=None, reference=None):

        if not isinstance(model, Model):
            raise ValueError("Model is incompatible or inexistent")

        self.model = model
        self.objective = self.model.objective if objective is None else objective
        self.environmental_conditions = OrderedDict() if envcond is None else envcond
        self.constraints = OrderedDict() if constraints is None else constraints
        self.solver = solver
        self._essential_reactions = None
        self._essential_genes = None
        self._gene_to_reaction = None
        self._reference = reference
        self.__status_mapping = {
            'optimal': SStatus.OPTIMAL,
            'unbounded': SStatus.UNBOUNDED,
            'infeasible': SStatus.INFEASIBLE,
            'infeasible_or_unbounded': SStatus.INF_OR_UNB,
            'suboptimal': SStatus.SUBOPTIMAL,
            'unknown': SStatus.UNKNOWN
        }
        self.solver = solver
        self._reset_solver = ModelConstants.RESET_SOLVER
        self.reverse_sintax = []

    @property
    def reference(self):
        """The reference wild type reaction flux values.

        :returns: A dictionary of wild type reaction flux values.

        """
        if self._reference is None:
            self._reference = self.simulate(
                method=SimulationMethod.pFBA).fluxes
        return self._reference

    @property
    def essential_reactions(self, min_growth=0.01):
        """Essential reactions are those when knocked out enable a biomass flux value above a minimal growth defined as
        a percentage of the wild type growth.

        :param float min_growth: Minimal percentage of the wild type growth value. Default 0.01 (1%).
        :returns: A list of essential reactions.

        """
        if self._essential_reactions is not None:
            return self._essential_reactions
        wt_solution = self.simulate()
        wt_growth = wt_solution.objective_value
        reactions = self.reactions
        self._essential_reactions = []
        for rxn in reactions:
            res = self.simulate(constraints={rxn: 0})
            if res:
                if (res.status == SStatus.OPTIMAL and res.objective_value < wt_growth * min_growth) \
                        or res.status == SStatus.INFEASIBLE:
                    self._essential_reactions.append(rxn)
        return self._essential_reactions

    @property
    def essential_genes(self, min_growth=0.01):
        """Essential genes are those when deleted enable a biomass flux value above a minimal growth defined as a
        percentage of the wild type growth.

        :param float min_growth: Minimal percentage of the wild type growth value. Default 0.01 (1%).
        :returns: A list of essential genes.

        """
        if self._essential_genes is not None:
            return self._essential_genes
        self._essential_genes = []
        wt_solution = self.simulate()
        wt_growth = wt_solution.objective_value
        genes = self.genes
        for gene in genes:
            active_genes = set(self.genes) - set([gene])
            active_reactions = self.evaluate_gprs(active_genes)
            inactive_reactions = set(self.reactions) - set(active_reactions)
            gr_constraints = {rxn: 0 for rxn in inactive_reactions}
            res = self.simulate(constraints=gr_constraints)
            if res:
                if (res.status == SStatus.OPTIMAL and res.objective_value < wt_growth * min_growth) \
                        or res.status == SStatus.INFEASIBLE:
                    self._essential_genes.append(gene)
        return self._essential_genes

    def evaluate_gprs(self, active_genes):
        """Returns the list of active reactions for a given list of active genes.

        :param list active_genes: List of genes identifiers.
        :returns: A list of active reaction identifiers.

        """
        active_reactions = []
        for r_id in self.reactions:
            reaction = self.model.reactions.get_by_id(r_id)
            if reaction.gene_reaction_rule:
                if evaluate_expression_tree(str(reaction.gene_reaction_rule), active_genes):
                    active_reactions.append(r_id)
            else:
                active_reactions.append(r_id)
        return active_reactions

    def get_uptake_reactions(self):
        """
        :returns: The list of uptake reactions.
        """
        drains = self.get_drains()
        rxns = [r for r in drains if self.model.reactions.get_by_id(r).reversibility
                or ((self.model.reactions.get_by_id(r).lower_bound is None
                    or self.model.reactions.get_by_id(r).lower_bound < 0)
                    and len(self.model.reactions.get_by_id(r).reactants) > 0)
                or ((self.model.reactions.get_by_id(r).upper_bound is None
                    or self.model.reactions.get_by_id(r).upper_bound > 0)
                    and len(self.model.reactions.get_by_id(r).products) > 0)
                ]
        return rxns

    def reverse_reaction(self, reaction_id):
        """
        Identify if a reaction is reversible and returns the reverse reaction if it is the case.

        :param reaction_id: A reaction identifier.
        :returns: A reaction identifier or None.

        """
        rxn = self.model.reactions.get_by_id(reaction_id)
        reactions = self.reactions
        if rxn.lower_bound < 0:
            return reaction_id
        else:
            rev = rxn.reverse_id
            if rev in reactions:
                return rev
        return None

    def gene_reactions(self):
        """
        :returns: A map of genes to reactions.
        """
        if not self._gene_to_reaction:
            gr = OrderedDict()
            for rxn_id in self.reactions:
                rxn = self.model.reactions.get_by_id(rxn_id)
                genes = rxn.genes
                for g in genes:
                    if g in gr.keys():
                        gr[g.id].append(rxn_id)
                    else:
                        gr[g.id] = [rxn_id]
            self._gene_to_reaction = gr
        return self._gene_to_reaction

    def get_reactions_for_genes(self, genes):
        """
        Returns the list of reactions catalysed by a list of genes

        :param list genes: A list of gene IDs.
        :returns: A list of reaction identifiers.

        """
        if not self._gene_to_reaction:
            self.gene_reactions()
        reactions = []
        for gene in genes:
            if gene in self._gene_to_reaction.keys():
                reactions.extend(self._gene_to_reaction[gene])
        return reactions

    def get_reaction_metabolites(self, reaction):
        '''
        Returns all metabolites of a given reaction.

        :param reaction: reaction (str)
        :return: metabolites (dict)

        '''
        return self.model.reactions.get_by_id(reaction).metabolites

    def is_reactant(self, reaction, metabolite):
        '''
        Returns if a metabolite is reactant into a given reaction.

        :param reaction: reaction (str)
        :param metabolite: metabolite (str)
        :return: bool

        '''

        return self.model.metabolites.get_by_id(metabolite) in self.model.reactions.get_by_id(reaction).reactants

    def is_product(self, reaction, metabolite):
        '''
        Returns if a metabolite is product into a given reaction.

        :param reaction: reaction (str)
        :param metabolite: metabolite (str)
        :return: bool

        '''
        return self.model.metabolites.get_by_id(metabolite) in self.model.reactions.get_by_id(reaction).products

    def get_metabolite_reactions(self, metabolite):

        return [reaction.id for reaction in self.model.metabolites.get_by_id(metabolite).reactions]

    def get_S(self):

        """
        Returns the S matrix as a numpy array

        :returns:

        S matrix, np.array

        """
        S = np.zeros((len(self.metabolites), len(self.reactions)))

        # usually there are more reactions than metabolites, but is far more easy to build S from reactions
        m_ids = {metabolite: i for i, metabolite in enumerate(self.model.metabolites)}

        for i, reaction in enumerate(self.model.reactions):
            for metabolite, coef in reaction.metabolites.items():
                S[m_ids[metabolite], i] = coef

        return S

    def get_reaction_bounds(self, reaction):
        """
        Returns the bounds for a given reaction.

        :param reaction: str, reaction ID
        :return: lb(s), ub(s), tuple

        """

        if reaction in self.constraints:
            lb, ub = self.constraints[reaction]
        elif reaction in self.environmental_conditions:
            lb, ub = self.environmental_conditions[reaction]
        else:
            lb, ub = self.model.reactions.get_by_id(reaction).bounds

        return lb if lb > -np.inf else -999999, ub if ub < np.inf else 999999

    def get_bounds(self):
        """
        Returns the whole set of lower and upper bounds as numpy arrays

        :returns: lb(s), ub(s), tuple of lists

        """

        lbs, ubs = list(zip(*[self.get_reaction_bounds(reaction) for reaction in self.reactions]))

        return list(lbs), list(ubs)

    def get_boundary_reaction(self, metabolite):

        """
        Finds the boundary reaction associated with an extracellular metabolite.
        If none is found, None is returned

        :param metabolite: str, metabolite ID
        :returns: reaction, str or None

        """

        for reaction in self.model.metabolites.get_by_id(metabolite).reactions:

            if reaction.id in self.model.exchanges \
                    or reaction.id in self.medium \
                    or reaction.boundary \
                    or (len(reaction.products) == 0):
                return reaction.id
        return None
    
    def find_bounds(self):
        """
        Return the median upper and lower bound of the metabolic model.
        Bounds can vary from model to model. Cobrapy defaults to (-1000, 1000).
        """
        lower_bounds = np.asarray([rxn.lower_bound for rxn in self.model.reactions], dtype=float)
        upper_bounds = np.asarray([rxn.upper_bound for rxn in self.model.reactions], dtype=float)
        lower_bound = np.nanmedian(lower_bounds[lower_bounds != 0.0])
        upper_bound = np.nanmedian(upper_bounds[upper_bounds != 0.0])
        if np.isnan(lower_bound):
            LOGGER.warning("Could not identify a median lower bound.")
            lower_bound = -1000.0
        if np.isnan(upper_bound):
            LOGGER.warning("Could not identify a median upper bound.")
            upper_bound = 1000.0
        return lower_bound, upper_bound
        
    def find_unconstrained_reactions(self):
        """Return list of reactions that are not constrained at all."""
        lower_bound, upper_bound = self.find_bounds()
        return [
            rxn
            for rxn in self.model.reactions
            if rxn.lower_bound <= lower_bound and rxn.upper_bound >= upper_bound
        ]

    def get_objective(self):
        """
        :returns: The model objective.

        """
        from cobra.util.solver import linear_reaction_coefficients
        return list(map(lambda x: x.id, linear_reaction_coefficients(self.model).keys()))

    # The simulator
    def simulate(self, objective=None, method=SimulationMethod.FBA, maximize=True,
                 constraints=None, reference=None, scalefactor=None):
        '''
        Simulates a phenotype when applying a set constraints using the specified method.

        :param dic objective: The simulation objective. If none, the model objective is considered.
        :param method: The SimulationMethod (FBA, pFBA, lMOMA, etc ...)
        :param boolean maximize: The optimization direction.
        :param dic constraints: A dictionary of constraints to be applied to the model.
        :param dic reference: A dictionary of reaction flux values.
        :param float scalefactor: A positive scaling factor for the solver. Default None.

        '''

        if not objective:
            objective = self.objective
        elif isinstance(objective, dict) and len(objective) > 0:
            objective = next(iter(objective.keys()))

        simul_constraints = OrderedDict()
        if constraints:
            simul_constraints.update(constraints)
        if self.constraints:
            simul_constraints.update(self.constraints)
        if self.environmental_conditions:
            simul_constraints.update(self.environmental_conditions)

        with self.model as model:
            model.objective = objective
            for rxn in list(simul_constraints.keys()):
                reac = model.reactions.get_by_id(rxn)
                # constraints defined as a tuple (lower_bound, upper_bound) or as a single float value
                if isinstance(simul_constraints.get(rxn), tuple):
                    reac.bounds = (simul_constraints.get(
                        rxn)[0], simul_constraints.get(rxn)[1])
                else:
                    reac.bounds = (simul_constraints.get(
                        rxn), simul_constraints.get(rxn))
            objective_sense = 'maximize' if maximize else 'minimize'
            if method == SimulationMethod.FBA:
                solution = model.optimize(objective_sense=objective_sense)
            elif method == SimulationMethod.pFBA:
                # make fraction_of_optimum a configurable parameter?
                solution = pfba(model)
            elif method == SimulationMethod.lMOMA or method == SimulationMethod.MOMA:
                s = None
                if not maximize:
                    s = model.optimize(objective_sense=objective_sense)
                linear = True if method == SimulationMethod.lMOMA else False
                solution = moma(model, solution=s, linear=linear)
            elif method == SimulationMethod.ROOM:
                solution = room(model)
            # Special case in which only the simulation context is required without any simulation result
            elif method == SimulationMethod.NONE:
                solution = Solution(None, 'unknown', None)
                pass
            else:
                raise Exception(
                    "Unknown method to perform the simulation.")

        status = self.__status_mapping[solution.status]

        result = SimulationResult(model, solution.objective_value, fluxes=solution.fluxes.to_dict(OrderedDict),
                                  status=status, envcond=self.environmental_conditions,
                                  model_constraints=self.constraints,
                                  simul_constraints=constraints,
                                  maximize=maximize)
        return result

    def FVA(self, obj_frac=0.9, reactions=None, constraints=None, loopless=False, internal=None, solver=None):
        """ Flux Variability Analysis (FVA).

        :param model: An instance of a constraint-based model.
        :param float obj_frac: The minimum fraction of the maximum growth rate (default 0.9). Requires that the objective value is at least the fraction times maximum objective value. A value of 0.85 for instance means that the objective has to be at least at 85% percent of its maximum.
        :param list reactions: List of reactions to analyze (default: all).
        :param dic constraints: Additional constraints (optional).
        :param boolean loopless: Run looplessFBA internally (very slow) (default: false).
        :param list internal: List of internal reactions for looplessFBA (optional).
        :param solver: A pre-instantiated solver instance (optional).
        :returns: A dictionary of flux variation ranges.

        """
        from cobra.flux_analysis.variability import flux_variability_analysis

        with self.model as model:

            if constraints:
                for rxn in list(constraints.keys()):
                    reac = model.reactions.get_by_id(rxn)
                    if isinstance(constraints.get(rxn), tuple):
                        reac.bounds = (constraints.get(
                            rxn)[0], constraints.get(rxn)[1])
                    else:
                        reac.bounds = (constraints.get(
                            rxn), constraints.get(rxn))

            if self.environmental_conditions:
                for rxn in list(self.environmental_conditions.keys()):
                    reac = model.reactions.get_by_id(rxn)
                    if isinstance(self.environmental_conditions.get(rxn), tuple):
                        reac.bounds = (self.environmental_conditions.get(
                            rxn)[0], self.environmental_conditions.get(rxn)[1])
                    else:
                        reac.bounds = (self.environmental_conditions.get(
                            rxn), self.environmental_conditions.get(rxn))
                
            df = flux_variability_analysis(
                model, reaction_list=reactions, loopless=loopless, fraction_of_optimum=obj_frac)

        variability = {}
        for r_id in reactions:
            variability[r_id] = [
                float(df.loc[r_id][0]), float(df.loc[r_id][1])]
        return variability

    def set_objective(self, reaction):
        self.model.objective = reaction


class GeckoSimulation(Simulation):
    """
    Simulator for geckopy.gecko.GeckoModel
    """

    def __init__(self, model: GeckoModel, objective=None, envcond=None, constraints=None,  solver=None, reference=None):
        super(GeckoSimulation, self).__init__(
            model, objective, envcond, constraints, solver, reference)
        self._essential_proteins = None
        self._protein_rev_reactions = None

    @property
    def proteins(self):
        return list(self.model.proteins)

    def essential_proteins(self, protein_prefix, min_growth=0.01):
        if self._essential_proteins is not None:
            return self._essential_proteins
        wt_solution = self.simulate()
        wt_growth = wt_solution.objective_value
        self._essential_proteins = []
        proteins = self.model.proteins
        for p in proteins:
            rxn = "{}{}".format(protein_prefix, p)
            res = self.simulate(constraints={rxn: 0})
            if res:
                if (res.status == SStatus.OPTIMAL and res.objective_value < wt_growth * min_growth) or \
                        res.status == SStatus.INFEASIBLE:
                    self._essential_proteins.append(rxn)
        return self._essential_proteins

    def protein_reactions(self, protein):
        """
        Returns the list of reactions associated to a protein.
        """
        reactions = []
        for r_id, rxn in self.model.reactions.items():
            lsub = rxn.reactants
            for m in lsub:
                if protein in m:
                    reactions.append(r_id)
        return reactions

    def reverse_reaction(self, reaction_id):
        """
        Identify if a reaction is reversible and returns the reverse reaction if it is the case.

        :param reaction_id: A reaction identifier.
        :returns: The reverse reaction identifier if exists or None.

        """
        f, d = zip(*self.protein_rev_reactions.values())
        if reaction_id in f:
            return d[f.index(reaction_id)]
        elif reaction_id in d:
            return f[d.index(reaction_id)]
        else:
            return None

    @property
    def protein_rev_reactions(self):
        """
        Pairs of reverse reactions associated with a protein

        :returns: A dictionary which identifies for each protein (key) the list of reversible reactions pairs.

        """
        if not self._protein_rev_reactions:
            proteins = self.model.proteins
            reactions = self.reactions
            in_sub = {}
            for p in proteins:
                sub = []
                for r_id in reactions:
                    rxn = self.model.reactions.get_by_id(r_id)
                    lsub = rxn.reactants
                    for m in lsub:
                        if p in m.id:
                            sub.append(r_id)
                in_sub[p] = sub
            pairs = {}
            for k, s in in_sub.items():
                revs = [r for r in s if '_REV' in r]
                if len(revs) > 0:
                    for r in revs:
                        la = [a for a in s if r.replace('_REV', '') == a]
                        la.append(r)
                        if len(la) == 2:
                            if k in pairs.keys():
                                pairs[k].append((la[0], la[1]))
                            else:
                                pairs[k] = [(la[0], la[1])]
            self._protein_rev_reactions = pairs
        return self._protein_rev_reactions



from mewpy.simulation import get_container, get_simulator


def load_sbml_container(filename, flavor='reframed'):
    if flavor == 'reframed':
        from reframed.io.sbml import load_cbmodel
        model = load_cbmodel(filename, flavor='cobra')
    elif flavor == 'cobra':
        from cobra.io import read_sbml_model
        model = read_sbml_model(filename)
    else:
        raise ValueError(f"{flavor} is not a recognized flavor")
    container = get_container(model)
    return container


def load_sbml_simulator(filename, flavor='reframed', envcond=None):
    if flavor == 'reframed':
        from reframed.io.sbml import load_cbmodel
        model = load_cbmodel(filename, flavor='cobra')
    elif flavor == 'cobra':
        from cobra.io import read_sbml_model
        model = read_sbml_model(filename)
    else:
        raise ValueError(f"{flavor} is not a recognized flavor")
    simul = get_simulator(model, envcond=envcond)
    return simul


def load_gecko_simulator(filename, flavor='reframed'):
    if flavor == 'reframed':
        from mewpy.model.gecko import GeckoModel
        model = GeckoModel(filename)
    elif flavor == 'cobra':
        from geckopy.gecko import GeckoModel
        model = GeckoModel(filename)
    else:
        raise ValueError(f"{flavor} is not a recognized flavor")
    simul = get_simulator(model)
    return simul



from warnings import warn
from mewpy.regulation import IntegratedModel
from mewpy.simulation import SStatus


class RFBAModel(IntegratedModel):

    """

    RFBAModel class object inherits everything from the IntegratedModel class.
    This object is the standard for simulating a RFBA model.
    First, a boolean regulatory model is solved synchronously
    It provides the simulate method where one can know all about the metabolic and regulatory analysis of RFBA

    """

    def __init__(self,
                 identifier,
                 name=None,
                 cbm_model=None,
                 cbm_simulation_interface=None,
                 targets=None,
                 regulators=None,
                 regulatory_interactions=None,
                 initial_state=None):

        # Integrated model
        super().__init__(identifier,
                         name,
                         cbm_model,
                         cbm_simulation_interface,
                         targets,
                         regulators,
                         regulatory_interactions,
                         initial_state)

    def get_current_state(self):

        """
        Retrieves the current metabolic and regulatory state

        :return: dict, for convenience returns dict of all regulatory variables where key is the id of the regulatory variable and value can be 0, 1 or float (reactions and metabolites predicates)
        """

        return {var.id: var.expression_coef for var in self.regulatory_variables_gen()}

    def __update_metabolic_genes_state__(self, state):

        # Updates only the state of the metabolic genes ignoring TFs and others

        for key, val in state.items():

            if key in self.metabolic_regulatory_genes:
                self.metabolic_regulatory_genes[key].expression_coef = val

    def __update_reactions_state__(self, decode):

        # Updates the regulatory variables considered to be reactions and available in the model.
        # It does so for the cbm simulation. The absolute value is considered, as the flux predicates might not
        # consider reversibility

        if decode.status != SStatus.OPTIMAL and decode.status != SStatus.SUBOPTIMAL:

            for val in self.metabolic_regulatory_reactions.values():
                val.expression_coef = float(0.0)

        else:

            for val in self.metabolic_regulatory_reactions.values():
                val.expression_coef = abs(float(decode.fluxes[val.cbm_model_id]))

    def __update_metabolites_state__(self):

        # The next comment version seems to be more robust to actually capture the presence or absence of a
        # metabolite.
        # The current version is taken from the matlab version, which just considers extracellular metabolites and
        # the associated reaction lower bound. This can also be adjusted for intracellular following the same line of
        # thought

        for val in self.metabolic_regulatory_metabolites.values():
            reaction = self.cbm_simulation_interface.get_boundary_reaction(val.cbm_model_id)

            val.expression_coef = abs(float(self.cbm_simulation_interface.get_reaction_bounds(reaction)[0]))

    # def __update_metabolites_state__(self, decode):
    #
    #     if decode.status != SStatus.OPTIMAL and decode.status != SStatus.SUBOPTIMAL:
    #
    #         for val in self.metabolic_regulatory_metabolites.values():
    #             val.expression_coef = float(0.0)
    #
    #     else:
    #
    #         for val in self.metabolic_regulatory_metabolites.values():
    #
    #             reactions = self.cbm_simulation_interface.get_metabolite_reactions(val.cbm_model_id)
    #
    #             value = 0
    #             for reaction in reactions:
    #
    #                 if (self.cbm_simulation_interface.is_product(reaction, val.cbm_model_id) and
    #                         (float(decode.fluxes[reaction]) > float(1E-10))):
    #
    #                     if abs(float(decode.fluxes[reaction])) > value:
    #                         value = abs(float(decode.fluxes[reaction]))
    #
    #                 if (self.cbm_simulation_interface.is_reactant(reaction, val.cbm_model_id) and
    #                         (float(decode.fluxes[reaction]) < float(-1E-10))):
    #
    #                     if abs(float(decode.fluxes[reaction])) > value:
    #                         value = abs(float(decode.fluxes[reaction]))
    #
    #             val.expression_coef = value

    def solve_regulatory_model(self, state, silently=False):

        """
        Solves the boolean regulatory network for a given specific state.
        It also updates all targets having a valid regulatory interaction associated with it for the resulting state


        :param state: dict, key is the id of the regulatory variable while value can be 0, 1 or float (reactions and metabolites predicates)
        :param silently: bool, if True, the state (expression_coef) of the regulatory variables is not changed
        :return: dict, for convenience returns dict where key is the id of the target and value can be 0, 1 or float (reactions and metabolites predicates)
        """

        res = {}

        # solving regulatory rule by regulatory rule (synchronously though, as asynchronously would take too much time)
        # Each target has one and only one regulatory interaction
        # The regulatory model object only supports this strategy, though this can easily be altered. If so,
        # this solve method must be changed too
        for val in self.regulatory_interactions_gen():

            state_map = {arg: state[arg] if arg in state else self.initial_state[arg]
                         for arg in val.variables}

            sol = val.evaluate(state_map)
            if sol is None:
                continue

            if not silently:
                val.target.expression_coef = sol
            res[val.id] = sol

        return res

    def next_state(self, state):

        """
        Retrieves the next state for a given state. This is mandatory for now.
        In the future, if no state is provided the current state will be used

        Solves the boolean regulatory model using method solve_regulatory_model(state) and decodes the metabolic state
        for the given state


        :param state: dict, dict, key is the id of the regulatory variable while value can be 0, 1 or float (reactions and metabolites predicates)
        :return: dict, for convenience returns dict of all regulatory variables where key is the id of the regulatory variable and value can be 0, 1 or float (reactions and metabolites predicates)
        """

        # Regulatory Rules from the boolean network
        self.solve_regulatory_model(state)

        # Reactions and metabolites from active genes, reactions and metabolites in the initial state
        decode = self.simulate_cbm_model(constraints=self.decode_metabolic_state(state))
        self.__update_reactions_state__(decode)

        # IF THE UPDATE METABOLITES STATE IS CHANGED FOR ACCOUNTING DECODED SOLUTIONS SUCH AS THE REACTIONS,
        # UNCOMMENT THIS!!! Otherwise, just use it once in the simulate
        # self.__update_metabolites_state__()

        return self.get_current_state()

    def __simulate__(self):

        self._regulatory_solution = []

        # For safety, we want to store a new dict and not the pointers
        new_state = {key: val for key, val in self.initial_state_gen()}
        self.regulatory_solution.append(new_state)

        # Perhaps this will change. See next state
        self.__update_metabolites_state__()

        # solve using the initial state
        new_state = self.next_state(new_state)
        self.regulatory_solution.append(new_state)
        return new_state

    def simulate(self, objective=None, maximize=True, method=None):

        """
        RFBA model one-step simulation (pseudo steady-state).

        First, the boolean regulatory model is solved for the initial state. If the initial state has not yet been
        set, the state of all regulatory variables is considered zero.

        At the same time, the metabolic state is also decoded from the initial state of all metabolic genes (once
        again, the metabolic genes state is zero if the initial state is not altered)

        The result of the boolean regulatory model updates the state of all targets, while the result of the
        metabolic state decoding updates the state of all reactions and metabolites that are also regulatory variables.

        Then, this new state is used to decode the final metabolic state using the resulting metabolic-regulatory state.

        Objective and constraint-based model analysis method (pFBA, FBA, ...) can be altered here reversibly.

        :param objective: None or dict, objective for the optimization. None for the current objective
        :param maximize: bool, direction for the optimization.
        :param method: None or SimulationMethod, constraint-based model analysis method
        :return: SimulationResult object, it contains the metabolic and regulatory states too.
        
        
        Solutions are also stored under solution property
        """

        self.objective = objective
        self.maximize = maximize
        self.cbm_simulation_method = method

        self.__simulate__()

        fba_sol = self.simulate_cbm_model(constraints=self.decode_metabolic_state(self.regulatory_solution[-1]))
        fba_sol.regulatory_solution = self.regulatory_solution[-1]
        self._solution = fba_sol

        self.__populate_solution__()

        return self._solution

    def dynamic_simulate(self, objective=None, maximize=True, method=None, iter_limit=1000):

        """
        RFBA model dynamic simulation (until the metabolic-regulatory state is reached).

        First, the boolean regulatory model is solved for the initial state. If the initial state has not yet been
        set, the state of all regulatory variables is considered zero.

        At the same time, the metabolic state is also decoded from the initial state of all metabolic genes (once
        again, the metabolic genes state is zero if the initial state is not altered)

        The result of the boolean regulatory model updates the state of all targets, while the result of the
        metabolic state decoding updates the state of all reactions and metabolites that are also regulatory variables.

        Then, this new resulting metabolic-regulatory state is used to solve again the boolean regulatory model and
        decode again the metabolic state towards the retrieval of a new regulatory and metabolic states.

        This cycle is iterated until a given state is repeated, the so called metabolic-regulatory steady-state.
        Hence, dynamic RFBA is based on finding the cyclic attractors of the metabolic-regulatory networks
        Alternatively, an iteration limit may be reached.

        Finally, all states between the cyclic attractor are used for decoding final metabolic states using
        the resulting metabolic-regulatory state. Thus, a solution/simulation is obtained for each mid-state in the
        cyclic attractor of the metabolic-regulatory networks

        Objective and constraint-based model analysis method (pFBA, FBA, ...) can be altered here reversibly.

        

        :param objective: None or dict, objective for the optimization. None for the current objective
        :param maximize: bool, direction for the optimization.
        :param method: None or SimulationMethod, constraint-based model analysis method
        :param iter_limit: int, iteration limit
        :return: list, list of SimulationResult objects containing the metabolic and regulatory states too.
        
        Solutions are also stored under solution property.
        
        """

        self.objective = objective
        self.maximize = maximize
        self.cbm_simulation_method = method

        new_state = self.__simulate__()

        i = 1
        sol = 1
        steady_state = False

        while not steady_state:

            # Updating state upon state. See next state for further detail
            new_state = self.next_state(new_state)
            self.regulatory_solution.append(new_state)

            n_sols = len(self.regulatory_solution) - 1

            for sol in range(n_sols):

                # A reaction can have different flux values between two solutions, though the state remains the same.
                # Thus, non-zero flux stands for ON state while zero flux stands for OFF state

                replica_v = []

                for key, val in self.regulatory_solution[sol].items():

                    sol_val = self.regulatory_solution[n_sols][key]

                    if key in self.metabolic_regulatory_reactions or key in self.metabolic_regulatory_metabolites:

                        if abs(val) > 1E-10:
                            val = True
                        else:
                            val = False

                        if abs(sol_val) > 1E-10:
                            sol_val = True
                        else:
                            sol_val = False

                    replica_v.append(val == sol_val)

                is_replica = all(replica_v)

                if is_replica:
                    steady_state = True

                    break

            if i < iter_limit:
                i += 1
            else:

                def iteration_limit(message):
                    warn(message, UserWarning, stacklevel=2)

                iteration_limit("Iteration limit reached")
                steady_state = True

        # The solution comprehends and FBA simulation between the first and second similar regulatory and metabolic
        # states using the corresponding regulatory states to switch on or off the associated reactions
        self._solution = []
        n_sols = len(self.regulatory_solution)
        for i in range(sol, n_sols):
            fba_sol = self.simulate_cbm_model(constraints=self.decode_metabolic_state(self.regulatory_solution[i]))
            fba_sol.regulatory_solution = self.regulatory_solution[i]
            self.solution.append(fba_sol)

        self.__populate_solution__()

        return self.solution



from mewpy.utils.parsing import build_tree, BooleanEvaluator, Boolean, GeneEvaluator, Arithmetic, ArithmeticEvaluator


def test_0():
    # aritmetic example
    t = build_tree(" 1 + 2 + 3 + ( 2 * 2 )", Arithmetic)
    res = t.evaluate(ArithmeticEvaluator.f_operand, ArithmeticEvaluator.f_operator)
    print(t, ' = ', res)


def test_1():
    # boolean example with conditions
    expression = "( (Lrp AND NOT (leu_L_e_>0)) OR NOT(((GlnG AND GlnB AND GlnD) AND RpoN) AND ((glu_L_e_>0) OR (arg_L_e_>0) OR (asp_L_e_>0) OR (his_L_e_>0) OR (pro_L_e_>0) )))"
    # expression = " (x > 0 or C or B) and not ph == 5 "
    t = build_tree(expression, Boolean)

    print(expression, ' <<==>> ', t)
    t.print_node()
    # list of true expressions
    true_list = ['GlnG']
    # dict of variables values
    v = {'leu_L_e_': 1, 'glu_L_e_': 7, "arg_L_e_": 0.5, "asp_L_e_": 2, "his_L_e_": 0, "pro_L_e_": 0}
    # propositions in the list are evaluated as True and the remaining are False
    evaluator = BooleanEvaluator(true_list, v)
    res = t.evaluate(evaluator.f_operand, evaluator.f_operator)
    print('evaluation: ', res)
    print('True list:', true_list)
    print('variables:', v)
    print('operands: ', t.get_operands())
    print('conditions: ', t.get_conditions())


def test_2():
    from urllib.request import urlretrieve
    from cobra.io import read_sbml_model
    import random

    path, _ = urlretrieve('http://bigg.ucsd.edu/static/models/RECON1.xml')
    model = read_sbml_model(path)
    ogpr = model.reactions.ATPS4m.gene_name_reaction_rule
    gpr = ogpr
    print(gpr)
    t = build_tree(gpr, Boolean)
    print(t)
    genes = list(t.get_operands())
    print("GENES:\n", genes)
    print("Evaluations:")
    evaluator = BooleanEvaluator(genes)
    res = t.evaluate(evaluator.f_operand, evaluator.f_operator)
    print(evaluator.true_list, " ==> ", res)
    for _ in range(20):
        l = []
        n = random.randint(0, len(genes))
        for _ in range(n):
            i = random.randint(0, len(genes) - 1)
            l.append(genes[i])
        evaluator.set_true_list(l)
        res = t.evaluate(evaluator.f_operand, evaluator.f_operator)
        print(evaluator.true_list, " ==> ", res)


def test_3():
    # Gene OU example
    gpr = "((G_YIL043C and G_YMR015C and G_YNL111C) or (G_YKL150W and G_YMR015C and G_YNL111C))"
    genes = {'G_YER091C': 0, 'G_YMR105C': 0.03125, 'G_YNL117W': 0.5, 'G_YNL111C': 0.125, 'G_YJR158W': 0.0625,
             'G_YLR355C': 0.5}
    operators = (lambda x, y: min(x, y), lambda x, y: max(x, y))
    evaluator = GeneEvaluator(genes, operators[0], operators[1])
    tree = build_tree(gpr, Boolean)
    tree.print_node()
    lv = tree.evaluate(evaluator.f_operand, evaluator.f_operator)
    print(f"{gpr}\n{genes}\nlevel:{lv}\n")


if __name__ == '__main__':
    test_1()
    # print()
    # test_2()
    # print()
    # test_3()



from .problem import AbstractKOProblem, AbstractOUProblem
from mewpy.simulation.kinetic import KineticSimulation
from collections import OrderedDict


class KineticKOProblem(AbstractKOProblem):

    def __init__(self, model, fevaluation=None, **kwargs):
        super(KineticKOProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        kinetic_parameters = kwargs.get('kparam', None)
        tSteps = kwargs.get('tSteps', None)
        self.kinetic_sim = KineticSimulation(model, parameters=kinetic_parameters, tSteps=tSteps)

    def _build_target_list(self):
        target = set(self.model.reactions.keys())
        if self.non_target is not None:
            target = target - set(self.non_target)
        self._trg_list = list(target)

    def decode(self, candidate):
        factors = OrderedDict([(self.target_list[idx], 0) for idx in candidate])
        return factors

    def evaluate_solution(self, solution, decode=True):
        """
        overrides evaluate_solution
        """
        p = []
        factors = self.decode(solution) if decode else solution
        simulation_results = self.kinetic_sim.simulate(factors=factors)
        for f in self.fevaluation:
            p.append(f(simulation_results, solution))
        return p


class KineticOUProblem(AbstractOUProblem):

    def __init__(self, model, fevaluation=None, **kwargs):
        super(KineticOUProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        kinetic_parameters = kwargs.get('kparam', None)
        tSteps = kwargs.get('tSteps', None)
        self.kinetic_sim = KineticSimulation(model, parameters=kinetic_parameters, tSteps=tSteps)

    def _build_target_list(self):
        target = set(self.model.reactions.keys())
        if self.non_target is not None:
            target = target - set(self.non_target)
        self._trg_list = list(target)

    def decode(self, candidate):
        factors = {self.target_list[idx]: self.levels[lv_idx] for idx, lv_idx in candidate}
        return factors

    def evaluate_solution(self, solution, decode=True):
        """
        overrides evaluate_solution
        """
        p = []
        factors = self.decode(solution) if decode else solution
        simulation_results = self.kinetic_sim.simulate(factors=factors)
        for f in self.fevaluation:
            p.append(f(simulation_results, solution))
        return p
